import re
import rltk
from dataclasses import dataclass
from datetime import datetime, MINYEAR
from enum import Enum
from typing import Tuple, Optional

import fastnumbers
import ftfy
from dateutil.parser import parse as dt_parse, ParserError

from kgdata.wikidata.models import DataValue, QNode


class WikidataValueType(Enum):
    string = "string"
    time = "time"
    quantity = "quantity"
    mono_lingual_text = "monolingualtext"
    globe_coordinate = "globecoordinate"
    entity_id = "wikibase-entityid"


@dataclass
class DatetimeParsedRepr:
    year: Optional[int] = None
    month: Optional[int] = None
    day: Optional[int] = None
    hour: Optional[int] = None
    minute: Optional[int] = None
    second: Optional[int] = None

    def has_only_year(self):
        return self.year is not None and \
               self.month is None and \
               self.day is None and \
               self.hour is None and \
               self.minute is None and \
               self.second is None

    def first_day_of_year(self):
        return self.year is not None and \
               self.month == 1 and \
               self.day == 1 and \
               self.hour is None and \
               self.minute is None and \
               self.second is None


@dataclass
class ParsedTextRepr:
    origin: str
    normed_string: str
    number_string: str
    number: Optional[float]
    datetime: Optional[DatetimeParsedRepr]


class TextParser:
    def __init__(self):
        self.cache = {}
        self.default_dt = datetime(MINYEAR, 1, 1)
        self.default_dt2 = datetime(MINYEAR+3, 2, 28)
        self.number_chars = re.compile("[^0-9\.+-]")

    def parse(self, text: str) -> ParsedTextRepr:
        if text not in self.cache:
            self.cache[text] = self._parse(text)
        return self.cache[text]

    def _parse(self, text: str):
        try:
            dt = dt_parse(text, default=self.default_dt)
            year = dt.year
            month = dt.month
            day = dt.day

            if dt.year == self.default_dt.year or dt.month == self.default_dt.month or dt.day == self.default_dt.day:
                dt2 = dt_parse(text, default=self.default_dt2)

                if dt.year == self.default_dt.year and dt2.year == self.default_dt2.year:
                    year = None
                if dt.month == self.default_dt.month and dt2.month == self.default_dt2.month:
                    month = None
                if dt.day == self.default_dt.day and dt2.day == self.default_dt2.day:
                    day = None

            dt = DatetimeParsedRepr(
                year=year,
                month=month,
                day=day
            )
        except (ParserError, TypeError, OverflowError):
            dt = None

        number_string = self._parse_number_string(text)
        normed_string = self._norm_string(text)
        if fastnumbers.isfloat(number_string) and (len(number_string) / min(1, len(normed_string))) > 0.95:
            number = fastnumbers.fast_real(number_string, coerce=False)
        else:
            number = None
        return ParsedTextRepr(
            origin=text,
            normed_string=normed_string,
            number_string=number_string,
            number=number,
            datetime=dt
        )

    def _parse_number_string(self, text: str):
        num_string = self.number_chars.sub("", text)
        return num_string

    def _norm_string(self, text: str):
        return ftfy.fix_text(text).replace("\xa0", " ").strip()


class LiteralMatcher:
    literal_types = {
        WikidataValueType.string.value, WikidataValueType.time.value,
        WikidataValueType.quantity.value, WikidataValueType.mono_lingual_text.value,
        WikidataValueType.globe_coordinate.value}
    non_literal_types = {WikidataValueType.entity_id.value}

    @classmethod
    def string_test_exact(cls, p_val: DataValue, val: ParsedTextRepr) -> Tuple[bool, float]:
        if val.normed_string == p_val.value:
            return True, 1.0
        return False, 0.0

    @classmethod
    def string_test_fuzzy(cls, p_val: DataValue, val: ParsedTextRepr) -> Tuple[bool, float]:
        return cls.match_string(val.normed_string, p_val.value)

    @classmethod
    def globe_coordinate_test(cls, p_val: DataValue, val: ParsedTextRepr) -> Tuple[bool, float]:
        return False, 0.0

    @classmethod
    def time_test(cls, p_val: DataValue, val: ParsedTextRepr) -> Tuple[bool, float]:
        if val.datetime is None:
            return False, 0.0
        timestr = p_val.value['time']
        # there are two calendar:
        # gregorian calendar, and julian calendar (just 13 days behind gregorian)
        # https://www.timeanddate.com/calendar/julian-gregorian-switch.html#:~:text=13%20Days%20Behind%20Today,days%20behind%20the%20Gregorian%20calendar.
        # TODO: we need to consider a range for julian calendar
        assert p_val.value['calendarmodel'] in {'http://www.wikidata.org/entity/Q1985786',
                                                'http://www.wikidata.org/entity/Q1985727'}, p_val.value['calendarmodel']
        # TODO: handle timezone, before/after and precision
        # pass

        # parse timestring
        match = re.match(r"([-+]\d+)-(\d{2})-(\d{2})T(\d{2}):(\d{2}):(\d{2})Z", timestr)

        target_val = DatetimeParsedRepr(
            year=int(match.group(1)),
            month=int(match.group(2)),
            day=int(match.group(3)),
            hour=int(match.group(4)),
            minute=int(match.group(5)),
            second=int(match.group(6)),
        )
        if timestr[0] == '-':
            target_val.year = -target_val.year
        else:
            assert timestr[0] == '+'
        for p in ['year', 'month', 'day', 'hour', 'minute', 'second']:
            if getattr(target_val, p) == 0:
                setattr(target_val, p, None)

        if target_val == val.datetime:
            return True, 1.0
        if val.datetime.has_only_year() and target_val.year == val.datetime.year:
            return True, 0.8
        if val.datetime.first_day_of_year() and target_val.year == val.datetime.year:
            return True, 0.75
        return False, 0.0

    @classmethod
    def quantity_test(cls, p_val: DataValue, val: ParsedTextRepr) -> Tuple[bool, float]:
        if val.number is None:
            return False, 0.0

        target_val = float(p_val.value["amount"])
        if abs(target_val - val.number) < 1e-5:
            return True, 1.0
        if target_val == 0:
            diff_percentage = abs(target_val - val.number) / 1e-7
        else:
            diff_percentage = abs((target_val - val.number) / target_val)

        if diff_percentage < 0.05:
            # within 5%
            return True, 0.95 - diff_percentage
        # if diff_percentage < 0.1:
        #     return True, 0.9 - diff_percentage
        return False, 0.0

    @classmethod
    def mono_lingual_text_test_exact(cls, p_val: DataValue, val: ParsedTextRepr) -> Tuple[bool, float]:
        target_val = p_val.value['text']
        if val.normed_string == target_val:
            return True, 1.0
        return False, 0.0

    @classmethod
    def mono_lingual_text_test_fuzzy(cls, p_val: DataValue, val: ParsedTextRepr) -> Tuple[bool, float]:
        target_val = p_val.value['text']
        return cls.match_string(val.normed_string, target_val)

    @classmethod
    def entity_id_test_fuzzy(cls, qnode: QNode, val: ParsedTextRepr) -> Tuple[bool, float]:
        string = val.normed_string
        lst = [x.strip() for x in string.split(",")]
        for label in [qnode.label] + qnode.aliases:
            for item in lst:
                result = cls.match_string(label, item)
                if result[0] is True:
                    return result
        return False, 0.0

    @classmethod
    def match_string(cls, s1: str, s2: str):
        if s1 == s2:
            return True, 1.0
        # calculate edit distance
        distance = rltk.levenshtein_distance(s1, s2)
        if distance <= 1:
            return True, 0.95
        elif distance > 1 and distance / max(1, min(len(s1), len(s2))) < 0.03:
            return True, 0.85
        return False, 0.0




from collections import defaultdict
from operator import itemgetter
from grams.algorithm.semantic_graph import SGEdge, SemanticGraphConstructor
from typing import *
import sm.misc as M
import networkx as nx


class SemTab2020PostProcessing:
    instance = None

    def __init__(self):
        from sm_unk.config import HOME_DIR
        dataset_dir = HOME_DIR / "wikitable2wikidata/semtab2020"
        self.CPA_targets = defaultdict(list)
        self.CTA_targets = defaultdict(list)
        for r in M.deserialize_csv(dataset_dir / "CPA_Round4_targets.csv"):
            self.CPA_targets[r[0]].append((int(r[1]), int(r[2])))
        for r in M.deserialize_csv(dataset_dir / "CTA_Round4_targets.csv"):
            self.CTA_targets[r[0]].append(int(r[1]))
    
    @staticmethod
    def get_instance():
        if SemTab2020PostProcessing.instance is None:
            SemTab2020PostProcessing.instance = SemTab2020PostProcessing()
        return SemTab2020PostProcessing.instance

    def solve_post_process(self, table, sg, dg, pred_with_probs: Dict[Tuple[str, str, str], float], threshold):
        # get the steiner tree first
        ci2id = {
            unode.column: uid
            for uid, unode in sg.nodes(data='data')
            if unode.is_column
        }
        selected_edges = set()
        for sci, tci in self.CPA_targets[table.id]:
            uid = ci2id[sci]
            vid = ci2id[tci]

            if not sg.has_node(uid) or not sg.has_node(vid):
                continue
            # among those paths, select the one with best prob
            paths = list(nx.all_simple_edge_paths(sg, uid, vid, cutoff=2))
            if len(paths) == 0:
                continue
            path_probs = []
            for path in paths:
                path_prob = 0
                for uid, vid, eid in path:
                    edge: SGEdge = sg.edges[uid, vid, eid]['data']
                    path_prob += pred_with_probs[uid, vid, eid]
                path_probs.append(path_prob)
            
            select_path, select_path_prob = max(zip(paths, path_probs), key=itemgetter(1))
            selected_edges = selected_edges.union(select_path)
        
        new_sg = SemanticGraphConstructor.get_sg_subgraph(sg, selected_edges)
        return new_sg
        
    def remove_dangling_statement(self, sg: nx.MultiDiGraph):
        ids = set()
        for sid, s in list(sg.nodes(data='data')):
            if s.is_statement and (sg.in_degree(sid) == 0 or sg.out_degree(sid) == 0):
                ids.add(sid)
        for id in ids:
            sg.remove_node(id)


import copy
import shutil
from operator import itemgetter
from pathlib import Path

from typing import Tuple, List

import networkx as nx
from redis import Redis

"""
Setup the input & output for testing the Steiner Tree solver
"""


class Env:
    def __init__(self, cfg=None, dataset_dir=None, dirname="default", n_tables: int=None,
                 load_qnode_test: bool=True, auto_path_increment: bool = False):
        if dataset_dir is None:
            dataset_dir = HOME_DIR / "wikitable2wikidata/250tables"
        self.dataset_dir = dataset_dir
        self.cfg = cfg
        # whether we enforce inverse constraint apply on object column.. only happen on the semtab2020 dataset
        self.strict = not self.dataset_dir.name.startswith("semtab2020")
        self.force_inversion = self.dataset_dir.name.startswith("semtab2020")

        self.gold_models = get_input_data(self.dataset_dir, self.dataset_dir.name, only_curated=True, n_tables=n_tables)
        self.qnodes = get_qnodes(self.dataset_dir, test=load_qnode_test, n_hop=self.cfg.data_graph.max_n_hop + 1)
        self.wdclasses = WDClass.from_file(self.dataset_dir / "ontology", load_parent_closure=True)
        self.wdprops = WDProperty.from_file(load_parent_closure=True)
        self.wd_numprop_stats = WDQuantityPropertyStats.from_dir()

        kg_index_file = dataset_dir / "kg_index" / "object_index.2hop_transitive.pkl.gz"
        if kg_index_file.exists():
            self.kg_object_index = KGObjectIndex.deserialize(kg_index_file, verbose=True)
        else:
            index_qnode_ids = list(get_qnodes(dataset_dir, n_hop=1, no_wdclass=True).keys())
            self.kg_object_index = KGObjectIndex.from_qnodes(index_qnode_ids, self.qnodes, self.wdprops,
                                                        n_hop=2, traversal_option=TraversalOption.TransitiveOnly,
                                                        verbose=True)
            self.kg_object_index.serialize(kg_index_file)

        self.evaluator = Evaluation(self.qnodes, self.wdclasses, self.wdprops)
        self.workdir = self.dataset_dir / "dev" / dirname
        if auto_path_increment:
            self.workdir = Path(M.get_incremental_path(self.workdir))
        self.workdir.mkdir(exist_ok=True, parents=True)

        cache_fn = lambda fn: M.redis_cache_func(REDIS_CACHE_URL, namespace=f"{self.dataset_dir.name}.{fn.__name__}", instance_method=False)(fn)

        # self.get_semantic_models = cache_fn(self.get_semantic_models)
        self.get_best_semantic_model = cache_fn(self.get_best_semantic_model)
        self.get_data_graph = cache_fn(self.get_data_graph)
        self.get_init_sg = cache_fn(self.get_init_sg)
        self.get_oracle_sgs = cache_fn(self.get_oracle_sgs)
        self.cache_fn = cache_fn

    def get_semantic_models(self, table_index: int):
        sms, table = self.gold_models[table_index]
        return [equiv_sm for sm in sms for equiv_sm in self.evaluator.wd_smhelper.gen_equivalent_sm(sm, self.strict, self.force_inversion)]

    def get_best_semantic_model(self, table_index: int):
        """Get the best semantic model that yield the maximum recall with the built data graph"""
        resp = self.get_init_sg(table_index)
        sms = self.get_semantic_models(table_index)

        perf = [
            (i, self.evaluator.oracle_pruning_cpa(resp.table, sm, copy.deepcopy(resp.sg)))
            for i, sm in enumerate(sms)
        ]
        i, perf = max(perf, key=lambda x: x[1]['recall'])
        return sms[i]

    def get_data_graph(self, table_index: int):
        sms, table = self.gold_models[table_index]

        build_data_graph_option = getattr(BuildDGOption, self.cfg.data_graph.options[0])
        for op in self.cfg.data_graph.options[1:]:
            build_data_graph_option = build_data_graph_option | getattr(BuildDGOption, op)

        dg = build_data_graph(table, self.qnodes, self.wdprops, self.kg_object_index, max_n_hop=self.cfg.data_graph.max_n_hop, options=build_data_graph_option)
        return dg

    def get_init_sg(self, table_index: int):
        sms, table = self.gold_models[table_index]
        dg = self.get_data_graph(table_index)
        constructor = SemanticGraphConstructor([
            SemanticGraphConstructor.init_sg,
            SemanticGraphConstructor.calculate_link_frequency,
            SemanticGraphConstructor.prune_sg_redundant_entity,
        ], self.qnodes, self.wdclasses, self.wdprops)
        resp = constructor.run(table, dg, debug=False)
        return resp

    def get_oracle_sgs(self, table_index: int):
        resp = self.get_init_sg(table_index)
        sms = self.get_semantic_models(table_index)
        perf = [
            self.evaluator.oracle_pruning_cpa(resp.table, sm, copy.deepcopy(resp.sg))
            for sm in sms
        ]
        best_recall = max([x['recall'] for x in perf])
        perf = [p for p in perf if p['recall'] == best_recall]

        oracle_sgs = []
        for p in perf:
            cpa_pred_sm: O.SemanticModel = p['cpa_pred_sm']

            # TODO: with new update from oracle_pruning_cpa, it's now also returned the semantic graph so we shouldn't to this. but needs to test before removing this code
            # have to exploit the fact that oracle_pruning_cpa function create a semantic model that reuse
            # the id in the original SG
            edges = [
                # convert uri back to predicate id to get the eid
                (edge.source, edge.target, self.evaluator.wd_smhelper.get_prop_id(edge.abs_uri))
                for edge in cpa_pred_sm.iter_edges()
            ]
            oracle_sg = SemanticGraphConstructor.get_sg_subgraph(resp.sg, edges)
            oracle_sgs.append(oracle_sg)
        return oracle_sgs

    def eval_cpa(self, table_index: int, st: nx.MultiDiGraph):
        """Evaluate the CPA performance"""
        table = self.gold_models[table_index][1]
        equiv_sms = self.get_semantic_models(table_index)
        cpa_sm, cpa_resp = max([
            (equiv_sm, self.evaluator.cpa(table, equiv_sm, st, False))
            for equiv_sm in equiv_sms
        ], key=lambda x: x[1]['f1'] if x[1] is not None else 0.0)
        if cpa_resp is None:
            result = (-1, -1, -1)
        else:
            result = (cpa_resp['precision'], cpa_resp['recall'], cpa_resp['f1'])
        return result, cpa_sm

    def eval_steiner_tree(self, table_index: int, st: nx.MultiDiGraph, oracle_sgs: List[nx.MultiDiGraph]=None):
        """Evaluate the Steiner Tree algorithm. Note this function is completely different
        with evaluate the correctness of the final semantic model (simply because it doesn't include
        all possible relationships).
        """
        if oracle_sgs is None:
            oracle_sgs = self.get_oracle_sgs(table_index)
        oracle_sms = [self.evaluator.convert_sg_for_cpa(sg) for sg in oracle_sgs]
        predict_sm = self.evaluator.convert_sg_for_cpa(st)
        best_resp = (-1, -1, -1)
        best_oracle_sg = None

        for (oracle_sm, oracle_sg) in zip(oracle_sms, oracle_sgs):
            resp = sm_metrics.precision_recall_f1(oracle_sm, predict_sm, self.evaluator.prop_score_fn)
            precision, recall, f1 = [resp[k] for k in ['precision', 'recall', 'f1']]

            if f1 > best_resp[-1]:
                best_resp = (precision, recall, f1)
                best_oracle_sg = oracle_sg

        return best_resp, best_oracle_sg

    def clear_viz_dir(self):
        if (self.workdir / "viz").exists():
            shutil.rmtree(self.workdir / "viz")
        (self.workdir / "viz").mkdir(exist_ok=True)

    def viz_dg(self, dg, id):
        viz_dg(dg, self.qnodes, self.wdprops, self.workdir / "viz", id)

    def viz_sg(self, sg, id):
        viz_sg(sg, self.qnodes, self.wdclasses, self.wdprops, self.workdir / "viz", id)

    def viz_sm(self, sm, id):
        self.evaluator.wd_smhelper.viz_sm(sm, self.workdir / "viz", id)

    def sg_to_sm(self, table_index, sg):
        """Convert semantic graph into semantic model for visualization purpose"""
        sm = O.SemanticModel()
        for uid, udata in sg.nodes(data=True):
            u: SGNode = udata['data']
            if u.is_statement:
                sm.add_node(
                    O.ClassNode(u.id, SemanticGraphConstructor.STATEMENT_URI,
                                SemanticGraphConstructor.STATEMENT_REL_URI, False, "Statement"))
            elif u.is_value:
                if u.is_entity_value:
                    sm.add_node(O.LiteralNode(u.id, self.evaluator.wd_smhelper.get_qnode_uri(u.qnode_id), readable_label=u.label))
                elif u.is_literal_value:
                    sm.add_node(O.LiteralNode(u.id, u.label, u.label))
            else:
                assert u.is_column
                sm.add_node(O.DataNode(u.id, u.column, self.gold_models[table_index][1].table.get_column_by_index(u.column)))

        for uid, vid, eid, edata in sg.edges(data=True, keys=True):
            e: SGEdge = edata['data']
            sm.add_edge(
                O.Edge(uid, vid, self.evaluator.wd_smhelper.get_prop_uri(e.predicate), f"p:{e.predicate}", False))

        return sm


import os
from dataclasses import dataclass
from operator import itemgetter
from pathlib import Path
from typing import Any, Dict, Set, Tuple, Union

import networkx as nx
from omegaconf import OmegaConf
from rdflib import RDFS

import grams.inputs as I
import sm.misc as M
import sm.outputs as O
from grams.algorithm.data_graph import build_data_graph, BuildDGOption
from grams.algorithm.kg_index import TraversalOption, KGObjectIndex
from grams.algorithm.psl_solver import PSLSteinerTreeSolver
from grams.algorithm.semantic_graph import (
    SemanticGraphConstructor,
)
from grams.algorithm.sm_wikidata import WikidataSemanticModelHelper
from grams.config import DEFAULT_CONFIG, ROOT_DIR
from kgdata.wikidata.db import (
    get_qnode_db,
    get_wdprop_db,
    get_wdclass_db,
    query_wikidata_entities,
)
from kgdata.wikidata.models import QNode, WDProperty, WDClass, WDQuantityPropertyStats


@dataclass
class Annotation:
    sm: O.SemanticModel
    # data graph
    dg: nx.MultiDiGraph
    # semantic graph
    sg: nx.MultiDiGraph
    # predicted semantic graph where incorrect relations are removed
    pred_sg: nx.MultiDiGraph
    # predicted column types
    pred_cta: Dict[int, Dict[str, float]]


class GRAMS:
    """Implementation of GRAMS. The main method is `annotate`"""

    def __init__(self, data_dir: Union[Path, str], cfg=None, proxy: bool = True):
        self.timer = M.Timer()
        self.cfg = cfg if cfg is None else DEFAULT_CONFIG
        self.is_proxy_db = proxy

        with self.timer.watch("init grams db"):
            read_only = not proxy
            self.qnodes = get_qnode_db(
                os.path.join(data_dir, "qnodes.db"),
                compression=True,
                read_only=read_only,
                proxy=proxy,
                is_singleton=True,
            )
            self.wdclasses = get_wdclass_db(
                os.path.join(data_dir, "wdclasses.db"),
                compression=False,
                read_only=read_only,
                proxy=proxy,
                is_singleton=True,
            )
            self.wdprops = get_wdprop_db(
                os.path.join(data_dir, "wdprops.db"),
                compression=False,
                read_only=read_only,
                proxy=proxy,
                is_singleton=True,
            )
            self.wd_numprop_stats = WDQuantityPropertyStats.from_dir(
                os.path.join(data_dir, "quantity_prop_stats")
            )

        self.build_dg_option = getattr(BuildDGOption, cfg.data_graph.options[0])
        for op in cfg.data_graph.options[1:]:
            self.build_dg_option = self.build_dg_option | getattr(BuildDGOption, op)

    def annotate(self, table: I.LinkedTable, verbose: bool = False) -> Annotation:
        """Annotate a linked table"""
        qnode_ids = {
            link.entity_id
            for rlinks in table.links
            for links in rlinks
            for link in links
            if link.entity_id is not None
        }
        qnode_ids.update(
            (
                candidate.entity_id
                for rlinks in table.links
                for links in rlinks
                for link in links
                for candidate in link.candidates
            )
        )
        if table.context.page_entity_id is not None:
            qnode_ids.add(table.context.page_entity_id)

        with self.timer.watch("retrieving qnodes"):
            qnodes = self.get_entities(qnode_ids, n_hop=2, verbose=verbose)
        wdclasses = self.wdclasses.cache_dict()
        wdprops = self.wdprops.cache_dict()

        if len(qnode_ids) != len(qnodes):
            nonexistent_qnode_ids = qnode_ids.difference(qnodes.keys())
            table.remove_nonexistent_entities(nonexistent_qnode_ids)

        with self.timer.watch("build kg object index"):
            kg_object_index = KGObjectIndex.from_qnodes(
                list(qnodes.keys()),
                qnodes,
                wdprops,
                n_hop=self.cfg.data_graph.max_n_hop,
                traversal_option=TraversalOption.TransitiveOnly,
            )

        with self.timer.watch("build dg & sg"):
            dg = build_data_graph(
                table,
                qnodes,
                wdprops,
                kg_object_index,
                max_n_hop=self.cfg.data_graph.max_n_hop,
                options=self.build_dg_option,
            )
            constructor = SemanticGraphConstructor(
                [
                    SemanticGraphConstructor.init_sg,
                ],
                qnodes,
                wdclasses,
                wdprops,
            )
            sg = constructor.run(table, dg, debug=False).sg

        with self.timer.watch("run inference"):
            psl_solver = PSLSteinerTreeSolver(
                qnodes,
                wdclasses,
                wdprops,
                self.wd_numprop_stats,
                disable_rules=set(self.cfg.psl.disable_rules),
                sim_fn=None,
                # cache_dir=outdir / f"{override_psl_cachedir}cache/psl",
                postprocessing_method=self.cfg.psl.postprocessing,
                enable_logging=self.cfg.psl.enable_logging,
            )
            pred_sg, pred_cta = psl_solver.run(
                dict(table=table, semanticgraph=sg, datagraph=dg)
            )
            pred_cta = {
                int(ci.replace("column-", "")): classes
                for ci, classes in pred_cta.items()
            }
            cta = {
                ci: max(classes.items(), key=itemgetter(1))[0]
                for ci, classes in pred_cta.items()
            }

        sm_helper = WikidataSemanticModelHelper(qnodes, wdclasses, wdprops)
        sm = sm_helper.create_sm(table, pred_sg, cta)
        sm = sm_helper.minify_sm(sm)
        return Annotation(sm=sm, dg=dg, sg=sg, pred_sg=pred_sg, pred_cta=cta)

    def get_entities(
        self, qnode_ids: Set[str], n_hop: int = 1, verbose: bool = False
    ) -> Dict[str, QNode]:
        assert n_hop <= 2
        batch_size = 30
        qnodes: Dict[str, QNode] = {}
        for qnode_id in qnode_ids:
            qnode = self.qnodes.get(qnode_id, None)
            if qnode is not None:
                qnodes[qnode_id] = qnode

        if self.is_proxy_db:
            missing_qnode_ids = [
                qnode_id
                for qnode_id in qnode_ids
                if qnode_id not in qnodes and not self.qnodes.does_not_exist(qnode_id)
            ]
            if len(missing_qnode_ids) > 0:
                resp = M.parallel_map(
                    query_wikidata_entities,
                    [
                        missing_qnode_ids[i : i + batch_size]
                        for i in range(0, len(missing_qnode_ids), batch_size)
                    ],
                    show_progress=verbose,
                    progress_desc=f"query wikidata for get missing entities in hop 1",
                    is_parallel=True,
                )
                for odict in resp:
                    for k, v in odict.items():
                        qnodes[k] = v
                        self.qnodes[k] = v

        if n_hop > 1:
            next_qnode_ids = set()
            for qnode in qnodes.values():
                for p, stmts in qnode.props.items():
                    for stmt in stmts:
                        if stmt.value.is_qnode():
                            next_qnode_ids.add(stmt.value.as_entity_id())
                        for qvals in stmt.qualifiers.values():
                            next_qnode_ids = next_qnode_ids.union(
                                qval.as_entity_id() for qval in qvals if qval.is_qnode()
                            )
            next_qnode_ids = list(next_qnode_ids.difference(qnodes.keys()))
            for qnode_id in next_qnode_ids:
                qnode = self.qnodes.get(qnode_id, None)
                if qnode is not None:
                    qnodes[qnode_id] = qnode

            if self.is_proxy_db:
                next_qnode_ids = [
                    qnode_id
                    for qnode_id in next_qnode_ids
                    if qnode_id not in qnodes
                    and not self.qnodes.does_not_exist(qnode_id)
                ]
                if len(next_qnode_ids) > 0:
                    resp = M.parallel_map(
                        query_wikidata_entities,
                        [
                            next_qnode_ids[i : i + batch_size]
                            for i in range(0, len(next_qnode_ids), batch_size)
                        ],
                        show_progress=verbose,
                        progress_desc=f"query wikidata for get missing entities in hop {n_hop}",
                        is_parallel=True,
                    )
                    for odict in resp:
                        for k, v in odict.items():
                            qnodes[k] = v
                            self.qnodes[k] = v
        return qnodes



import copy
import re
from collections import defaultdict
from dataclasses import dataclass, field
from enum import Enum, Flag, auto
from functools import cmp_to_key
from itertools import chain
from typing import Dict, Optional, Union, List, Set, Tuple, NamedTuple, Any

import networkx as nx
from loguru import logger
from tqdm import tqdm
from typing_extensions import TypedDict

from grams.algorithm.kg_index import KGObjectIndex
from grams.algorithm.literal_match import WikidataValueType, TextParser, LiteralMatcher
from grams.inputs.linked_table import LinkedTable
from kgdata.wikidata.models import QNode, DataValue, WDProperty, WDClass
import sm.misc as M
from sm.misc.graph import viz_graph


@dataclass(frozen=True)
class Span:
    start: int
    end: int

    @property
    def length(self):
        return self.end - self.start


@dataclass
class CellNode:
    id: str
    value: str
    column: int
    row: int
    qnode_ids: List[str]
    qnodes_span: Dict[str, List[Span]]

    @property
    def is_cell(self):
        return True

    @property
    def is_literal_value(self):
        return False

    @property
    def is_entity_value(self):
        return False

    @property
    def is_statement(self):
        return False


@dataclass
class ContextSpan:
    # the text of the context
    text: str
    span: Span

    def get_text_span(self):
        return self.text[self.span.start : self.span.end]


@dataclass
class LiteralValueNode:
    id: str
    value: DataValue
    # not none if it is appear in the context
    context_span: Optional[ContextSpan]

    @property
    def is_cell(self):
        return False

    @property
    def is_literal_value(self):
        return True

    @property
    def is_entity_value(self):
        return False

    @property
    def is_statement(self):
        return False

    @property
    def is_context(self):
        return self.context_span is not None


@dataclass
class EntityValueNode:
    id: str
    qnode_id: str
    # not none if it is appear in the context
    context_span: Optional[ContextSpan]

    @property
    def is_cell(self):
        return False

    @property
    def is_literal_value(self):
        return False

    @property
    def is_entity_value(self):
        return True

    @property
    def is_statement(self):
        return False

    @property
    def is_context(self):
        return self.context_span is not None


EdgeFlowSource = NamedTuple("EdgeFlowSource", [("source_id", str), ("edge_id", str)])
EdgeFlowTarget = NamedTuple("EdgeFlowTarget", [("target_id", str), ("edge_id", str)])


class LinkGenMethod(Enum):
    """Methods to generate a link"""

    # this statement is generated by exact matching from the link
    FromWikidataLink = "from_wikidata_link"
    FromLiteralMatchingFunc = "from_literal_matching_function"


@dataclass
class FlowProvenance:
    """Contain information regarding how this relationship/flow has been generated (typically coming from the matching algorithm)"""

    # method that
    gen_method: LinkGenMethod
    gen_method_arg: Any
    prob: float

    def merge(self, another: "FlowProvenance") -> Union["FlowProvenance", None]:
        """Try to merge the provenance, if we cannot merge them, just return None"""
        if self.gen_method != another.gen_method:
            return None
        if self.gen_method == LinkGenMethod.FromWikidataLink:
            return self
        if self.gen_method == LinkGenMethod.FromLiteralMatchingFunc:
            if self.gen_method_arg == another.gen_method_arg:
                if self.prob > another.prob:
                    return self
                return another
            return None
        raise M.UnreachableError()

    @staticmethod
    def merge_lst(
        lst1: List["FlowProvenance"], lst2: List["FlowProvenance"]
    ) -> List["FlowProvenance"]:
        """Assume that items within each list (lst1 & lst2) are not mergeable"""
        if len(lst1) == 0:
            return lst2
        elif len(lst2) == 0:
            return lst1

        lst = copy.copy(lst1)
        for item in lst2:
            for i in range(len(lst1)):
                resp = lst[i].merge(item)
                if resp is None:
                    # cannot merge them
                    lst.append(item)
                    break
                else:
                    # merge lst[i] & item
                    lst[i] = resp
                    break
        return lst


@dataclass
class StatementNode:
    id: str
    # id of the qnode that contains the statement
    qnode_id: str
    # predicate of the statement
    predicate: str
    # whether this statement actually exist in KG
    is_in_kg: bool

    # recording which link in the source is connected to the target.
    forward_flow: Dict[EdgeFlowSource, Set[EdgeFlowTarget]] = field(
        default_factory=dict
    )
    reversed_flow: Dict[EdgeFlowTarget, Set[EdgeFlowSource]] = field(
        default_factory=dict
    )
    flow: Dict[Tuple[EdgeFlowSource, EdgeFlowTarget], List[FlowProvenance]] = field(
        default_factory=dict
    )

    @property
    def value(self):
        return self.id

    @property
    def is_cell(self):
        return False

    @property
    def is_literal_value(self):
        return False

    @property
    def is_entity_value(self):
        return False

    @property
    def is_statement(self):
        return True

    def track_provenance(
        self,
        source_flow: EdgeFlowSource,
        target_flow: EdgeFlowTarget,
        provenances: List[FlowProvenance],
    ):
        ptr = self.forward_flow
        if source_flow not in ptr:
            ptr[source_flow] = set()
        ptr[source_flow].add(target_flow)

        ptr = self.reversed_flow
        if target_flow not in ptr:
            ptr[target_flow] = set()
        ptr[target_flow].add(source_flow)

        # TODO: merge the provenance if we can
        if (source_flow, target_flow) not in self.flow:
            self.flow[source_flow, target_flow] = provenances
        else:
            self.flow[source_flow, target_flow] = FlowProvenance.merge_lst(
                self.flow[source_flow, target_flow], provenances
            )

    def untrack_source_flow(self, source_flow: EdgeFlowSource):
        for target_flow in self.forward_flow.pop(source_flow):
            self.flow.pop((source_flow, target_flow))
            self.reversed_flow[target_flow].remove(source_flow)
            if len(self.reversed_flow[target_flow]) == 0:
                self.reversed_flow.pop(target_flow)

    def untrack_target_flow(self, target_flow: EdgeFlowTarget):
        for source_flow in self.reversed_flow.pop(target_flow):
            self.flow.pop((source_flow, target_flow))
            self.forward_flow[source_flow].remove(target_flow)
            if len(self.forward_flow[source_flow]) == 0:
                self.forward_flow.pop(source_flow)

    def untrack_flow(self, source_flow: EdgeFlowSource, target_flow: EdgeFlowTarget):
        self.flow.pop((source_flow, target_flow))
        self.forward_flow[source_flow].remove(target_flow)
        self.reversed_flow[target_flow].remove(source_flow)

        if len(self.forward_flow[source_flow]) == 0:
            self.forward_flow.pop(source_flow)
        if len(self.reversed_flow[target_flow]) == 0:
            self.reversed_flow.pop(target_flow)

    def has_source_flow(self, source_flow):
        return source_flow in self.forward_flow

    def has_target_flow(self, target_flow):
        return target_flow in self.reversed_flow

    def has_flow(self, source_flow, target_flow):
        return (source_flow, target_flow) in self.flow

    def iter_source_flow(self, target_flow: EdgeFlowTarget):
        for source_flow in self.reversed_flow[target_flow]:
            yield source_flow, self.flow[source_flow, target_flow]

    def iter_target_flow(self, source_flow: EdgeFlowSource):
        for target_flow in self.forward_flow[source_flow]:
            yield target_flow, self.flow[source_flow, target_flow]

    def get_provenance(self, source_flow: EdgeFlowSource, target_flow: EdgeFlowTarget):
        return self.flow[source_flow, target_flow]

    def get_provenance_by_edge(self, inedge: "DGEdge", outedge: "DGEdge"):
        return self.flow[
            EdgeFlowSource(inedge.source, inedge.predicate),
            EdgeFlowTarget(outedge.target, outedge.predicate),
        ]

    def is_same_flow(self, inedge: "DGEdge", outedge: "DGEdge") -> bool:
        return (
            EdgeFlowSource(inedge.source, inedge.predicate),
            EdgeFlowTarget(outedge.target, outedge.predicate),
        ) in self.flow


DGNode = Union[CellNode, LiteralValueNode, EntityValueNode, StatementNode]


@dataclass
class DGStatementID:
    qnode_id: str
    predicate: str
    statement_index: int

    def get_id(self):
        return f"stmt:{self.qnode_id}-{self.predicate}-{self.statement_index}"

    @staticmethod
    def parse_id(id: str) -> "DGStatementID":
        m = re.match(r"stmt:([^-]+)-([^-]+)-([^-]+)", id)
        assert m is not None
        qnode_id, predicate, stmt_index = [m.group(i) for i in range(1, 4)]
        stmt_index = int(stmt_index)
        return DGStatementID(qnode_id, predicate, stmt_index)


@dataclass
class DGPathNodeStatement:
    qnode_id: str
    predicate: str
    statement_index: int
    # how this DGPath statement has been matched
    provenance: FlowProvenance

    def get_id(self):
        return DGStatementID(
            self.qnode_id, self.predicate, self.statement_index
        ).get_id()

    # ############# METHODs to construct DGPathNodeStatement from provenance ##############
    @staticmethod
    def from_FromWikidataLink(qnode_id, predicate, stmt_index):
        return DGPathNodeStatement(
            qnode_id,
            predicate,
            stmt_index,
            FlowProvenance(LinkGenMethod.FromWikidataLink, None, 1.0),
        )

    @staticmethod
    def from_FromLiteralMatchingFunc(qnode_id, predicate, stmt_index, fn_args, prob):
        return DGPathNodeStatement(
            qnode_id,
            predicate,
            stmt_index,
            FlowProvenance(LinkGenMethod.FromLiteralMatchingFunc, fn_args, prob),
        )


@dataclass
class DGPathNodeQNode:
    qnode_id: str

    def get_id(self):
        return f"ent:{self.qnode_id}"


@dataclass
class DGPathNodeLiteralValue:
    value: DataValue

    def get_id(self):
        return f"val:{self.value.to_string_repr()}"


@dataclass
class DGPathExistingNode:
    id: str

    def get_id(self):
        return self.id


DGPathNode = Union[
    DGPathNodeStatement, DGPathNodeQNode, DGPathNodeLiteralValue, DGPathExistingNode
]


@dataclass
class DGPathEdge:
    value: str
    is_qualifier: bool

    @staticmethod
    def p(value: str):
        return DGPathEdge(value, is_qualifier=False)

    @staticmethod
    def q(value: str):
        return DGPathEdge(value, is_qualifier=True)


@dataclass
class DGPath:
    # a sequence of path, is always
    sequence: List[Union[DGPathEdge, DGPathNode]] = field(default_factory=list)


@dataclass
class DGEdge:
    source: str
    target: str
    predicate: str
    is_qualifier: bool
    # deprecated, will be replaced by the information stored directly in the statement edge flow
    # paths: List[DGPath]
    is_inferred: Optional[bool] = None

    @staticmethod
    def can_link(source: DGNode, target: DGNode):
        """Test if there can be two links between the two nodes. Basically, it just enforce the constraint that no link
        cross different rows
        """
        if not source.is_cell or not target.is_cell:
            return True
        return source.row == target.row


@dataclass
class InferredNewProp:
    # information to identify the statement for the new prop (whether to reuse it, or create a new one)
    # the qnode of the new statement (the one will have this property)
    qnode_id: str
    # the new property we inferred
    new_prop: str
    # the value associated with the property (use it to compare if the statement exists) - the reason we use
    # the value instead of the whole statement is that sometimes, desire them to match the qualifier does not
    # sense, if the value exist, then some how the algorithm is already match it so it should be okay
    value: DataValue

    # the source nodes in the data graph
    source_id: str
    # the target node in the data graph that will contain the value of the prop, not the statement node
    target_id: str

    # edges that contain the qualifiers that we want to be copied to the new statement as well
    # the flow of the qualifier can be retrieve
    qualifier_edges: List[DGEdge]

    # flow provenance of the property.
    flow_provenances: List[FlowProvenance]


@dataclass
class InferredNewQualifier:
    # the statement node in the data graph that we will add the qualifier to
    statement_id: str
    # the qualifier we are going to add
    new_qualifier: str

    # source id (node that has the statement id)
    source_id: str
    # the property of the statement (the one that connect source and statement)
    property: str
    # the target node of the qualifier
    target_id: str

    # flow provenance of the qualifier
    flow_provenances: List[FlowProvenance]


class BuildDGOption(Flag):
    # disable kg search for the same entity
    DISABLE_SAME_ENT_SEARCH = auto()
    # removing unnecessary entities to keep the graph size reasonable
    PRUNING_REDUNDANT_ENT = auto()
    # disable path discovering using index
    NO_KG_INDEX = auto()


def kg_path_discovering(
    options: BuildDGOption,
    qnodes: Dict[str, QNode],
    kg_object_index: KGObjectIndex,
    dg: nx.MultiDiGraph,
    u: DGNode,
    v: DGNode,
    max_n_hop: int = 2,
    bidirection: bool = True,
):
    """Find all paths between two nodes in the graph.

    Parameters
    ----------
    options: options when building DG
    qnodes: dictionary of qnodes
    dg: the input data graph
    u: node
    v: node
    max_n_hop: hop start at 1 (current node), n_hop = 2 mean the path can go through an hidden node
    bidirection: set to false telling if we want to just find paths going from u -> v.
        This parameter just for re-using code, as we will call the function changing the order of u & v

    Returns
    -------
    """

    # noinspection PyShadowingNames
    def _path_object_search_v1(
        qnodes: Dict[str, QNode],
        kg_object_index: KGObjectIndex,
        source: QNode,
        target: QNode,
        max_n_hop=2,
    ):
        # max_n_hop = 2 mean we will find a path that go from source to target through an hidden node
        # hop start at 1 (current node)

        if options & BuildDGOption.DISABLE_SAME_ENT_SEARCH:
            if source.id == target.id:
                return []

        matches = []
        iter = source.props.items()

        for p, stmts in iter:
            if p == "P31":
                # no need to search in the instanceOf property, as the ontology is removed from the databased as they are huge
                continue

            for stmt_i, stmt in enumerate(stmts):
                # add the cell id so that we have different statement for different cell.
                has_stmt_value = False

                # to simplify the design, we do not consider a statement that its value do not exist in KQ
                # due to an error on KG
                if stmt.value.is_qnode():
                    if stmt.value.as_entity_id() not in qnodes:
                        # this can happen due to some of the qnodes is in the link, but is missing in the KG
                        # this is very rare so we can employ some check to make sure this is not due to
                        # our wikidata subset
                        is_error_in_kg = any(
                            any(
                                _s.value.is_qnode()
                                and _s.value.as_entity_id() in qnodes
                                for _s in _stmts
                            )
                            for _p, _stmts in source.props.items()
                        ) or stmt.value.as_entity_id().startswith("L")
                        if not is_error_in_kg:
                            raise Exception(
                                f"Missing qnodes in your KG subset: {stmt.value.as_entity_id()}"
                            )
                        continue

                if stmt.value.is_qnode():
                    # found by match entity
                    if stmt.value.as_entity_id() == target.id:
                        matches.append(
                            DGPath(
                                sequence=[
                                    DGPathEdge.p(p),
                                    DGPathNodeStatement.from_FromWikidataLink(
                                        source.id, p, stmt_i
                                    ),
                                    DGPathEdge.p(p),
                                ]
                            )
                        )
                        has_stmt_value = True
                    elif max_n_hop > 1:
                        stmt_value_qnode_id = stmt.value.as_entity_id()
                        if stmt_value_qnode_id.startswith("L"):
                            assert (
                                stmt_value_qnode_id not in qnodes
                            ), "The L nodes (lexical) is not in the Wikidata dump"
                            continue

                        for nextpath in _path_object_search(
                            qnodes,
                            kg_object_index,
                            qnodes[stmt_value_qnode_id],
                            target,
                            max_n_hop - 1,
                        ):
                            matches.append(
                                DGPath(
                                    sequence=[
                                        DGPathEdge.p(p),
                                        DGPathNodeStatement.from_FromWikidataLink(
                                            source.id, p, stmt_i
                                        ),
                                        DGPathEdge.p(p),
                                        DGPathNodeQNode(stmt_value_qnode_id),
                                    ]
                                    + nextpath.sequence
                                )
                            )
                            has_stmt_value = True

                for q, qvals in stmt.qualifiers.items():
                    for qval in qvals:
                        if qval.is_qnode():
                            if qval.as_entity_id() == target.id:
                                if not has_stmt_value:
                                    if stmt.value.is_qnode():
                                        pn_stmt_value = DGPathNodeQNode(
                                            stmt.value.as_entity_id()
                                        )
                                    else:
                                        pn_stmt_value = DGPathNodeLiteralValue(
                                            stmt.value
                                        )
                                    matches.append(
                                        DGPath(
                                            sequence=[
                                                DGPathEdge.p(p),
                                                DGPathNodeStatement.from_FromWikidataLink(
                                                    source.id, p, stmt_i
                                                ),
                                                DGPathEdge.p(p),
                                                pn_stmt_value,
                                            ]
                                        )
                                    )
                                    has_stmt_value = True

                                matches.append(
                                    DGPath(
                                        sequence=[
                                            DGPathEdge.p(p),
                                            DGPathNodeStatement.from_FromWikidataLink(
                                                source.id, p, stmt_i
                                            ),
                                            DGPathEdge.q(q),
                                        ]
                                    )
                                )
                            elif max_n_hop > 1:
                                qval_qnode_id = qval.as_entity_id()
                                if qval_qnode_id.startswith("L"):
                                    assert (
                                        qval_qnode_id not in qnodes
                                    ), "The L nodes (lexical) is not in the Wikidata dump"
                                    continue

                                if qval_qnode_id not in qnodes:
                                    # this can happen due to some of the qnodes is in the link, but is missing in the KG
                                    # this is very rare so we can employ some check to make sure this is not due to
                                    # our wikidata subset
                                    is_error_in_kg = any(
                                        any(
                                            _s.value.is_qnode()
                                            and _s.value.as_entity_id() in qnodes
                                            for _s in _stmts
                                        )
                                        for _p, _stmts in source.props.items()
                                    ) or qval_qnode_id.startswith("L")
                                    if not is_error_in_kg:
                                        raise Exception(
                                            f"Missing qnodes in your KG subset: {qval_qnode_id}"
                                        )
                                    continue

                                _n_matches = len(matches)
                                for nextpath in _path_object_search(
                                    qnodes,
                                    kg_object_index,
                                    qnodes[qval_qnode_id],
                                    target,
                                    max_n_hop - 1,
                                ):
                                    matches.append(
                                        DGPath(
                                            sequence=[
                                                DGPathEdge.p(p),
                                                DGPathNodeStatement.from_FromWikidataLink(
                                                    source.id, p, stmt_i
                                                ),
                                                DGPathEdge.q(q),
                                                DGPathNodeQNode(qval.as_entity_id()),
                                            ]
                                            + nextpath.sequence
                                        )
                                    )

                                if len(matches) > _n_matches and not has_stmt_value:
                                    if stmt.value.is_qnode():
                                        pn_stmt_value = DGPathNodeQNode(
                                            stmt.value.as_entity_id()
                                        )
                                    else:
                                        pn_stmt_value = DGPathNodeLiteralValue(
                                            stmt.value
                                        )
                                    matches.append(
                                        DGPath(
                                            sequence=[
                                                DGPathEdge.p(p),
                                                DGPathNodeStatement.from_FromWikidataLink(
                                                    source.id, p, stmt_i
                                                ),
                                                DGPathEdge.p(p),
                                                pn_stmt_value,
                                            ]
                                        )
                                    )
                                    has_stmt_value = True

        return matches

    def _path_object_search_v2(
        qnode: Dict[str, QNode],
        kg_object_index: KGObjectIndex,
        source: QNode,
        target: QNode,
        max_n_hop=2,
    ):
        # max_n_hop = 2 mean we will find a path that go from source to target through an hidden node
        # hop start at 1 (current node)

        if options & BuildDGOption.DISABLE_SAME_ENT_SEARCH:
            if source.id == target.id:
                return []

        matches = []
        for hop1_path in kg_object_index.iter_hop1_props(source.id, target.id):
            stmt_i = hop1_path.statement_index
            rel = hop1_path.relationship
            stmt = source.props[rel.prop][stmt_i]

            if len(rel.quals) > 0 and not rel.both:
                # the prop doesn't match, have to add it
                if stmt.value.is_qnode():
                    pn_stmt_value = DGPathNodeQNode(stmt.value.as_entity_id())
                else:
                    pn_stmt_value = DGPathNodeLiteralValue(stmt.value)
                matches.append(
                    DGPath(
                        sequence=[
                            DGPathEdge.p(rel.prop),
                            DGPathNodeStatement.from_FromWikidataLink(
                                source.id, rel.prop, stmt_i
                            ),
                            DGPathEdge.p(rel.prop),
                            pn_stmt_value,
                        ]
                    )
                )
            else:
                # the prop match
                matches.append(
                    DGPath(
                        sequence=[
                            DGPathEdge.p(rel.prop),
                            DGPathNodeStatement.from_FromWikidataLink(
                                source.id, rel.prop, stmt_i
                            ),
                            DGPathEdge.p(rel.prop),
                        ]
                    )
                )
            for qual in rel.quals:
                matches.append(
                    DGPath(
                        sequence=[
                            DGPathEdge.p(rel.prop),
                            DGPathNodeStatement.from_FromWikidataLink(
                                source.id, rel.prop, stmt_i
                            ),
                            DGPathEdge.q(qual),
                        ]
                    )
                )

        if max_n_hop > 1:
            # TODO: need to refactor the current method of searching for paths
            # as we may generate many duplicated segments
            for match_item in kg_object_index.iter_hop2_props(source.id, target.id):
                if match_item.qnode == target.id:
                    continue
                hop1_paths = match_item.hop1
                hop2_paths = match_item.hop2
                middle_qnode = qnodes[match_item.qnode]

                # we don't care about transitive here, so we do a cross product
                hop1_seqs = []
                for hop1_path in hop1_paths:
                    rel = hop1_path.relationship
                    stmt_i = hop1_path.statement_index
                    stmt = source.props[rel.prop][stmt_i]

                    if len(rel.quals) > 0 and not rel.both:
                        # we don't have the statement value yet, add it to the matches
                        # the prop doesn't match, have to add it, we don't worry about duplication
                        # as it is resolve during the merge provenance phase
                        if stmt.value.is_qnode():
                            pn_stmt_value = DGPathNodeQNode(stmt.value.as_entity_id())
                        else:
                            pn_stmt_value = DGPathNodeLiteralValue(stmt.value)
                        matches.append(
                            DGPath(
                                sequence=[
                                    DGPathEdge.p(rel.prop),
                                    DGPathNodeStatement.from_FromWikidataLink(
                                        source.id, rel.prop, stmt_i
                                    ),
                                    DGPathEdge.p(rel.prop),
                                    pn_stmt_value,
                                ]
                            )
                        )
                    else:
                        # add prop to the seqs that we need to expand next, and so stmt.value must be a qnode
                        # as it is the middle qnode
                        assert stmt.value.as_entity_id() == middle_qnode.id
                        hop1_seqs.append(
                            [
                                DGPathEdge.p(rel.prop),
                                DGPathNodeStatement.from_FromWikidataLink(
                                    source.id, rel.prop, stmt_i
                                ),
                                DGPathEdge.p(rel.prop),
                                DGPathNodeQNode(middle_qnode.id),
                            ]
                        )

                    for qual in rel.quals:
                        hop1_seqs.append(
                            [
                                DGPathEdge.p(rel.prop),
                                DGPathNodeStatement.from_FromWikidataLink(
                                    source.id, rel.prop, stmt_i
                                ),
                                DGPathEdge.p(qual),
                                DGPathNodeQNode(middle_qnode.id),
                            ]
                        )

                for hop2_path in hop2_paths:
                    rel = hop2_path.relationship
                    stmt_i = hop2_path.statement_index
                    stmt = middle_qnode.props[rel.prop][stmt_i]

                    if len(rel.quals) > 0 and not rel.both:
                        if stmt.value.is_qnode():
                            pn_stmt_value = DGPathNodeQNode(stmt.value.as_entity_id())
                        else:
                            pn_stmt_value = DGPathNodeLiteralValue(stmt.value)

                        hop2_seq = [
                            DGPathEdge.p(rel.prop),
                            DGPathNodeStatement.from_FromWikidataLink(
                                middle_qnode.id, rel.prop, stmt_i
                            ),
                            DGPathEdge.p(rel.prop),
                            pn_stmt_value,
                        ]
                        for hop1_seq in hop1_seqs:
                            matches.append(DGPath(sequence=hop1_seq + hop2_seq))
                    else:
                        hop2_seq = [
                            DGPathEdge.p(rel.prop),
                            DGPathNodeStatement.from_FromWikidataLink(
                                middle_qnode.id, rel.prop, stmt_i
                            ),
                            DGPathEdge.p(rel.prop),
                        ]
                        for hop1_seq in hop1_seqs:
                            matches.append(DGPath(sequence=hop1_seq + hop2_seq))

                    for qual in rel.quals:
                        hop2_seq = [
                            DGPathEdge.p(rel.prop),
                            DGPathNodeStatement.from_FromWikidataLink(
                                middle_qnode.id, rel.prop, stmt_i
                            ),
                            DGPathEdge.p(qual),
                        ]
                        for hop1_seq in hop1_seqs:
                            matches.append(DGPath(sequence=hop1_seq + hop2_seq))

        return matches

    if options & BuildDGOption.NO_KG_INDEX:
        _path_object_search = _path_object_search_v1
    else:
        _path_object_search = _path_object_search_v2

    textparser = TextParser()
    funcs = {
        WikidataValueType.string.value: [LiteralMatcher.string_test_exact],
        WikidataValueType.globe_coordinate.value: [
            LiteralMatcher.globe_coordinate_test
        ],
        WikidataValueType.time.value: [LiteralMatcher.time_test],
        WikidataValueType.quantity.value: [LiteralMatcher.quantity_test],
        WikidataValueType.mono_lingual_text.value: [
            LiteralMatcher.mono_lingual_text_test_exact
        ],
        WikidataValueType.entity_id.value: [],  # [LiteralMatcher.entity_id_test_fuzzy]
    }

    # noinspection PyUnusedLocal,PyShadowingNames
    def _path_value_search(qnodes: Dict[str, QNode], source: QNode, value):
        matches = []
        for p, stmts in source.props.items():
            if p == "P31":
                # no need to search in the instanceOf property, as the ontology is removed from the databased as they are huge
                continue

            for stmt_i, stmt in enumerate(stmts):
                has_stmt_value = False
                if stmt.value.type in LiteralMatcher.literal_types:
                    stmt_value = stmt.value
                else:
                    assert stmt.value.type in LiteralMatcher.non_literal_types
                    if not stmt.value.is_qnode():
                        # lexical
                        continue
                    if stmt.value.as_entity_id() not in qnodes:
                        # this can happen due to some of the qnodes is in the link, but is missing in the KG
                        # this is very rare so we can employ some check to make sure this is not due to
                        # our wikidata subset
                        is_error_in_kg = any(
                            any(
                                _s.value.is_qnode()
                                and _s.value.as_entity_id() in qnodes
                                for _s in _stmts
                            )
                            for _p, _stmts in source.props.items()
                        ) or stmt.value.as_entity_id().startswith("L")
                        if not is_error_in_kg:
                            raise Exception(
                                f"Missing qnodes in your KG subset: {stmt.value.as_entity_id()}"
                            )
                        continue
                    stmt_value = qnodes[stmt.value.as_entity_id()]
                for fn in funcs[stmt.value.type]:
                    match, confidence = fn(stmt_value, value)
                    if match:
                        matches.append(
                            DGPath(
                                sequence=[
                                    DGPathEdge.p(p),
                                    DGPathNodeStatement.from_FromLiteralMatchingFunc(
                                        source.id,
                                        p,
                                        stmt_i,
                                        {"func": fn.__name__, "value": stmt.value},
                                        confidence,
                                    ),
                                    DGPathEdge.p(p),
                                ]
                            )
                        )
                        has_stmt_value = True

                for q, qvals in stmt.qualifiers.items():
                    for qval in qvals:
                        if qval.type in LiteralMatcher.literal_types:
                            qval_value = qval
                        else:
                            assert qval.type in LiteralMatcher.non_literal_types
                            # TODO: fix me handle L not correct
                            if qval.value["id"].startswith("L") or qval.value[
                                "id"
                            ].startswith("P"):
                                continue
                            if qval.as_entity_id() not in qnodes:
                                # this can happen due to some of the qnodes is in the link, but is missing in the KG
                                # this is very rare so we can employ some check to make sure this is not due to
                                # our wikidata subset
                                is_error_in_kg = any(
                                    any(
                                        _s.value.is_qnode()
                                        and _s.value.as_entity_id() in qnodes
                                        for _s in _stmts
                                    )
                                    for _p, _stmts in source.props.items()
                                ) or not qval.as_entity_id().startswith("L")
                                if not is_error_in_kg:
                                    raise Exception(
                                        f"Missing qnodes in your KG subset: {qval.as_entity_id()}"
                                    )
                                continue
                            qval_value = qnodes[qval.as_entity_id()]
                        for fn in funcs[qval.type]:
                            match, confidence = fn(qval_value, value)
                            if match:
                                if not has_stmt_value:
                                    if stmt.value.is_qnode():
                                        pn_stmt_value = DGPathNodeQNode(
                                            stmt.value.as_entity_id()
                                        )
                                    else:
                                        pn_stmt_value = DGPathNodeLiteralValue(
                                            stmt.value
                                        )
                                    matches.append(
                                        DGPath(
                                            sequence=[
                                                DGPathEdge.p(p),
                                                DGPathNodeStatement.from_FromWikidataLink(
                                                    source.id, p, stmt_i
                                                ),
                                                DGPathEdge.p(p),
                                                pn_stmt_value,
                                            ]
                                        )
                                    )
                                    has_stmt_value = True

                                matches.append(
                                    DGPath(
                                        sequence=[
                                            DGPathEdge.p(p),
                                            DGPathNodeStatement.from_FromLiteralMatchingFunc(
                                                source.id,
                                                p,
                                                stmt_i,
                                                {"func": fn.__name__, "value": qval},
                                                confidence,
                                            ),
                                            DGPathEdge.q(q),
                                        ]
                                    )
                                )
        return matches

    paths = []

    # no qnode in the source
    u_qnodes = (
        [qnodes[u.qnode_id]]
        if u.is_entity_value
        else [qnodes[qnode_id] for qnode_id in u.qnode_ids]
    )
    if len(u_qnodes) == 0:
        if bidirection:
            return kg_path_discovering(
                options,
                qnodes,
                kg_object_index,
                dg,
                v,
                u,
                max_n_hop=max_n_hop,
                bidirection=False,
            )
        return []

    v_qnodes = (
        [qnodes[v.qnode_id]]
        if v.is_entity_value
        else [qnodes[qnode_id] for qnode_id in v.qnode_ids]
    )
    v_value = v.context_span.get_text_span() if v.is_entity_value else v.value
    for n1 in u_qnodes:
        # count = timer.start("object discovery")
        for n2 in v_qnodes:
            for newpath in _path_object_search(
                qnodes, kg_object_index, n1, n2, max_n_hop=max_n_hop
            ):
                newpath.sequence.insert(0, DGPathExistingNode(u.id))
                if isinstance(newpath.sequence[-1], DGPathEdge):
                    newpath.sequence.append(DGPathExistingNode(v.id))
                paths.append(newpath)
        # count.stop()

        # count = timer.start("literal discovery")
        for newpath in _path_value_search(qnodes, n1, textparser.parse(v_value)):
            newpath.sequence.insert(0, DGPathExistingNode(u.id))
            if isinstance(newpath.sequence[-1], DGPathEdge):
                newpath.sequence.append(DGPathExistingNode(v.id))
            paths.append(newpath)
        # count.stop()

    if bidirection:
        paths += kg_path_discovering(
            options,
            qnodes,
            kg_object_index,
            dg,
            v,
            u,
            max_n_hop=max_n_hop,
            bidirection=False,
        )

    return paths


class KGInference:
    def __init__(
        self,
        dg: nx.MultiDiGraph,
        qnodes: Dict[str, QNode],
        wdprops: Dict[str, WDProperty],
    ):
        # mapping from qnode id, and property to a list of statement values (with the corresponding DG statement node if exist)
        # the reason we want to include all is that we want to know if we need to add new statement value or reuse existing value
        self.subkg: Dict[
            Tuple[str, str], List[Tuple[Optional[StatementNode], DataValue]]
        ] = {}
        self.qnodes = qnodes
        self.wdprops = wdprops
        self.dg = dg

        for sid, sdata in dg.nodes(data=True):
            s: StatementNode = sdata["data"]
            if not s.is_statement:
                continue

            dgsid = DGStatementID.parse_id(sid)
            self._set_stmt_node(s.qnode_id, dgsid.predicate, dgsid.statement_index, s)

    def infer_subproperty(self):
        """Infer new properties via sub-property of (inheritance)"""
        properties = set()
        qualifiers = set()

        for uid, udata in self.dg.nodes(data=True):
            u = udata["data"]
            if u.is_statement:
                continue

            for sid, us_edges in self.dg[u.id].items():
                # add all links to the list of properties
                properties.update(us_edges.keys())

                for vid, sv_edges in self.dg[sid].items():
                    qualifiers.update(
                        (v_eid for v_eid in sv_edges.keys() if v_eid not in us_edges)
                    )

        parent_props = self._build_parent_map(properties)
        parent_qualifiers = self._build_parent_map(qualifiers)

        # list of new properties and qualifiers that we will infer
        new_props = []
        new_qualifiers = []

        for sid, sdata in self.dg.nodes(data=True):
            s: StatementNode = sdata["data"]
            if not s.is_statement:
                continue

            # parents of the statement
            parents = set()
            prop = s.predicate
            for uid, _, eid, edata in self.dg.in_edges(sid, data=True, keys=True):
                parents.add(uid)
                assert prop == edata["data"].predicate

            # children that are properties of the statement
            prop_children = set()
            # children that are qualifiers of the statement
            qualifier2children = defaultdict(set)
            for _, vid, eid, edata in self.dg.out_edges(sid, data=True, keys=True):
                e: DGEdge = edata["data"]
                if e.predicate == prop:
                    prop_children.add(vid)
                else:
                    qualifier2children[e.predicate].add(vid)

            stmt_index = DGStatementID.parse_id(sid).statement_index
            if prop in parent_props:
                for parent_prop in parent_props[prop]:
                    for vid in prop_children:
                        for source_flow, flow_provenances in s.iter_source_flow(
                            EdgeFlowTarget(vid, prop)
                        ):
                            new_prop = InferredNewProp(
                                qnode_id=s.qnode_id,
                                new_prop=parent_prop,
                                value=self._get_stmt_value(
                                    s.qnode_id, prop, stmt_index
                                ),
                                source_id=source_flow.source_id,
                                target_id=vid,
                                qualifier_edges=[],
                                flow_provenances=flow_provenances,
                            )
                            new_props.append(new_prop)

            for q, children in qualifier2children.items():
                for pq in parent_qualifiers.get(q, []):
                    for vid in children:
                        for source_flow, provenance in s.iter_source_flow(
                            EdgeFlowTarget(vid, q)
                        ):
                            new_qualifiers.append(
                                InferredNewQualifier(
                                    statement_id=sid,
                                    new_qualifier=pq,
                                    source_id=source_flow.source_id,
                                    property=source_flow.edge_id,
                                    target_id=vid,
                                    flow_provenances=provenance,
                                )
                            )

        self.add_inference(new_props, new_qualifiers)
        return self

    def kg_transitive_inference(self):
        """Infer new relationship based on the transitive property: a -> b -> c => a -> c"""

        # find the list of transitive properties in the graph
        transitive_props = set()
        for uid, vid, eid, edata in self.dg.edges(data=True, keys=True):
            prop = self.wdprops[eid]
            # transitive class
            if "Q18647515" in prop.instanceof:
                transitive_props.add(eid)

        # now start from node u, we find if there is another v connect to u via a transitive property, and another p connect
        # to v with the same property, we don't need to keep the chain going as even if it's longer, we will eventually loop
        # through all item in the chain by looping through nodes in the graph

        chains = []
        for uid, udata in self.dg.nodes(data=True):
            u: DGNode = udata["data"]
            if u.is_statement:
                continue

            for sid, us_edges in self.dg[uid].items():
                stmt: StatementNode = self.dg.nodes[sid]["data"]
                for trans_prop in transitive_props:
                    if trans_prop not in us_edges:
                        continue
                    us_edge: DGEdge = us_edges[trans_prop]["data"]
                    for vid, sv_edges in self.dg[sid].items():
                        if trans_prop not in sv_edges:
                            continue
                        sv_edge: DGEdge = sv_edges[trans_prop]["data"]
                        if not stmt.is_same_flow(us_edge, sv_edge):
                            # don't allow infer new link across rows
                            continue

                        for s2id, vs2_edges in self.dg[vid].items():
                            if trans_prop not in vs2_edges:
                                continue
                            vs2_edge: DGEdge = vs2_edges[trans_prop]["data"]
                            stmt2: StatementNode = self.dg.nodes[s2id]["data"]
                            for v2id, s2v2_edges in self.dg[s2id].items():
                                if trans_prop not in s2v2_edges:
                                    continue
                                s2v2_edge: DGEdge = s2v2_edges[trans_prop]["data"]
                                if not stmt2.is_same_flow(vs2_edge, s2v2_edge):
                                    # don't allow infer new link across rows
                                    continue

                                # we now record the chain
                                chains.append((us_edge, sv_edge, vs2_edge, s2v2_edge))

        # make sure that there is no duplication in the chains
        assert len(chains) == len(
            {
                (
                    us_edge.source,
                    us_edge.predicate,
                    us_edge.target,
                    sv_edge.source,
                    sv_edge.predicate,
                    sv_edge.target,
                    vs2_edge.source,
                    vs2_edge.predicate,
                    vs2_edge.target,
                    s2v2_edge.source,
                    s2v2_edge.predicate,
                    s2v2_edge.target,
                )
                for us_edge, sv_edge, vs2_edge, s2v2_edge in chains
            }
        )

        # generate new property, but qualifiers cannot inherit via transitive inference
        new_props = []
        for us_edge, sv_edge, vs2_edge, s2v2_edge in chains:
            trans_prop = us_edge.predicate
            stmt: StatementNode = self.dg.nodes[us_edge.target]["data"]
            stmt2: StatementNode = self.dg.nodes[vs2_edge.target]["data"]

            prop_value = self._get_stmt_value(
                stmt2.qnode_id,
                trans_prop,
                DGStatementID.parse_id(stmt2.id).statement_index,
            )
            # prop_value = stmt2.qnode.props[trans_prop][DGStatementID.parse_id(stmt2.id).statement_index].value

            # calculating the provenance of the new transitive link. however, I don't know in case we have multiple
            # provenance since it depends on both first leg and second leg. for now, we put an assertion to handle
            # only the case where we have two legs has the same provenance, which make the new transitive link has
            # the same provenance too.
            first_leg_provenances = stmt.get_provenance_by_edge(us_edge, sv_edge)
            second_leg_provenances = stmt2.get_provenance_by_edge(vs2_edge, s2v2_edge)
            assert (
                len(first_leg_provenances) == 1
                and len(second_leg_provenances) == 1
                and first_leg_provenances[0] == second_leg_provenances[0]
            )

            provenances = first_leg_provenances
            new_props.append(
                InferredNewProp(
                    qnode_id=stmt.qnode_id,
                    new_prop=trans_prop,
                    value=prop_value,
                    source_id=us_edge.source,
                    target_id=s2v2_edge.target,
                    qualifier_edges=[],
                    flow_provenances=provenances,
                )
            )

        self.add_inference(new_props, [])
        return self

    def add_inference(
        self,
        new_props: List[InferredNewProp],
        new_qualifiers: List[InferredNewQualifier],
    ):
        """After we run inference, we got a list of new properties and new qualifiers that can update using this function"""
        new_nodes = []
        new_edges = []

        # here we enforce the constraint that there is no cross links
        # between nodes in different rows, this happen because transitive
        # inference generate links cross rows
        new_props = [
            new_prop
            for new_prop in new_props
            if DGEdge.can_link(
                self.dg.nodes[new_prop.source_id]["data"],
                self.dg.nodes[new_prop.target_id]["data"],
            )
        ]
        new_qualifiers = [
            new_qualifier
            for new_qualifier in new_qualifiers
            if DGEdge.can_link(
                self.dg.nodes[new_qualifier.source_id]["data"],
                self.dg.nodes[new_qualifier.target_id]["data"],
            )
        ]

        for new_prop in new_props:
            stmt_exist = None
            prop = new_prop.new_prop

            # search for existing statement in the KG
            self._track_property(new_prop.qnode_id, prop)
            for sprime, value in self.subkg[new_prop.qnode_id, prop]:
                if sprime is not None and new_prop.value == value:
                    stmt_exist = sprime
                    break

            # if the statement exist, re-use it
            if stmt_exist is not None:
                sprime = stmt_exist
            else:
                for stmt_index, (sprime, value) in enumerate(
                    self.subkg[new_prop.qnode_id, prop]
                ):
                    if new_prop.value == value:
                        stmt_id = DGStatementID(
                            new_prop.qnode_id, prop, stmt_index
                        ).get_id()
                        sprime = StatementNode(
                            stmt_id, new_prop.qnode_id, prop, is_in_kg=True
                        )
                        self._set_stmt_node(new_prop.qnode_id, prop, stmt_index, sprime)
                        new_nodes.append(sprime)
                        break
                else:
                    stmt_index = self.get_next_available_stmt_index(
                        new_prop.qnode_id, prop
                    )
                    stmt_id = DGStatementID(
                        new_prop.qnode_id, prop, stmt_index
                    ).get_id()
                    sprime = StatementNode(
                        stmt_id, new_prop.qnode_id, prop, is_in_kg=False
                    )
                    self._add_stmt_value(
                        new_prop.qnode_id, prop, stmt_index, sprime, new_prop.value
                    )
                new_nodes.append(sprime)

            if not self.dg.has_edge(new_prop.source_id, sprime.id, key=prop):
                new_edges.append(
                    DGEdge(
                        source=new_prop.source_id,
                        target=sprime.id,
                        predicate=prop,
                        is_qualifier=False,
                        is_inferred=True,
                    )
                )
            if not self.dg.has_edge(sprime.id, new_prop.target_id, key=prop):
                new_edges.append(
                    DGEdge(
                        source=sprime.id,
                        target=new_prop.target_id,
                        predicate=prop,
                        is_qualifier=False,
                        is_inferred=True,
                    )
                )

            sprime.track_provenance(
                EdgeFlowSource(new_prop.source_id, prop),
                EdgeFlowTarget(new_prop.target_id, prop),
                new_prop.flow_provenances,
            )

            # TODO: we haven't add to add add qualifiers, so we assert we don't have any. fix me!
            assert len(new_prop.qualifier_edges) == 0
            # for qual_edge in new_prop.qualifier_edges:
            #     new_edges.append(DGEdge(source=stmt_id, target=qual_edge.target, predicate=qual_edge.predicate, is_qualifier=True, paths=[], is_inferred=True))

        for new_qual in new_qualifiers:
            stmt: StatementNode = self.dg.nodes[new_qual.statement_id]["data"]
            new_edges.append(
                DGEdge(
                    source=new_qual.statement_id,
                    target=new_qual.target_id,
                    predicate=new_qual.new_qualifier,
                    is_qualifier=True,
                    is_inferred=True,
                )
            )
            stmt.track_provenance(
                EdgeFlowSource(new_qual.source_id, new_qual.property),
                EdgeFlowTarget(new_qual.target_id, new_qual.new_qualifier),
                new_qual.flow_provenances,
            )

        for node in new_nodes:
            self.dg.add_node(node.id, data=node)
        for edge in new_edges:
            self.dg.add_edge(edge.source, edge.target, key=edge.predicate, data=edge)

    def _track_property(self, qnode_id: str, prop: str):
        """Ensure that the subkg has values of the qnode's property"""
        if (qnode_id, prop) not in self.subkg:
            lst = []
            for stmt_i, stmt in enumerate(self.qnodes[qnode_id].props.get(prop, [])):
                lst.append((None, stmt.value))
            self.subkg[qnode_id, prop] = lst

    def _get_stmt_value(self, qnode_id: str, prop: str, stmt_index: int):
        self._track_property(qnode_id, prop)
        return self.subkg[qnode_id, prop][stmt_index][1]

    def _set_stmt_node(
        self, qnode_id: str, prop: str, stmt_index: int, stmt: StatementNode
    ):
        self._track_property(qnode_id, prop)
        assert (
            self.subkg[qnode_id, prop][stmt_index][0] is None
        ), "Cannot override existing value in the KG"
        self.subkg[qnode_id, prop][stmt_index] = (
            stmt,
            self.subkg[qnode_id, prop][stmt_index][1],
        )

    def _add_stmt_value(
        self,
        qnode_id: str,
        prop: str,
        stmt_index: int,
        stmt: StatementNode,
        value: DataValue,
    ):
        self._track_property(qnode_id, prop)
        assert stmt_index == len(self.subkg[qnode_id, prop]), "Can only add new value"
        self.subkg[qnode_id, prop].append((stmt, value))

    def get_next_available_stmt_index(self, qnode_id: str, prop: str):
        return len(self.subkg[qnode_id, prop])

    def _build_parent_map(self, props: Set[str]):
        """Build a map from a property to its parents in the same list"""
        parent_props: Dict[str, List[str]] = {}
        for p1 in props:
            parent_props[p1] = []
            for p2 in props:
                if p1 == p2:
                    continue
                if p2 in self.wdprops[p1].parents_closure:
                    parent_props[p1].append(p2)
            if len(parent_props[p1]) == 0:
                parent_props.pop(p1)
        return parent_props


class DGPruning:
    NxDGEdgeAttr = TypedDict("NxDGEdgeAttr", data=DGEdge)
    NxDGEdge = Tuple[str, str, str, NxDGEdgeAttr]

    def __init__(self, dg: nx.MultiDiGraph):
        self.dg = dg

    def prune_hidden_entities(self):
        """Prune redundant KG entities, which added to the graph via KG discovering and from the context.

        **Step 1:**
        Let:
        - n be an entity node in DG.
        - v is a node connected from n via a property: LEG2: n -> p -> s -> p' -> v, and s does not have other property/qualifier rather than p'

        We made the following heuristics:
        * If there is no other node connect to n, then n is a root node and is from the context. We should not
        prune this node, so just skip it.
        * For all another node ui \in U that connects to n via the path: LEG1: ui -> pi -> s' -> pi' -> n, if there is always a better
        path LEG* between ui and v, then we can remove the path LEG2. U contains nodes in cells or context, otherwise
        ui will be an entity to entity that won't be in the final model anyway.
        Note: LEG* is better than LEG2 when it's shorter, also from wikidata link or if not, it must have better match confidence (i.e., better provenance)

        **Step 2:**
        Let n' be an entity node in DG that do not link to other nodes (v doesn't exist).
        We have the following heuristics:
        * If there is no other node connect to it, this is a standable node and should be removed
        * For any node ui that connects to n via the path: LEG1: ui -> pi -> s' -> pi' -> n. If s' doesn't have other properties/qualifiers,
        then we can remove LEG1.

        **Step 3:**
        Finally, if a node is standable, we should remove it.
        """
        # step 1: prune the second leg paths
        legprime: Dict[Tuple[str, str], FlowProvenance] = {}
        rm_legs: List[Tuple[str, EdgeFlowSource, EdgeFlowTarget]] = []
        for nid, ndata in self.dg.nodes(data=True):
            n: EntityValueNode = ndata["data"]
            if not n.is_entity_value:
                continue

            if self.dg.in_degree(nid) == 0:
                # no other node connect to it
                continue

            # get list of grandpa ui (only cells or values in the context), along with their paths to node n.
            grandpa = set()
            for gpid in self.iter_grand_parents(nid):
                gp = self.dg.nodes[gpid]["data"]
                if gp.is_cell or gp.is_context:
                    grandpa.add(gpid)

            for _, sid, ns_eid in self.dg.out_edges(nid, keys=True):
                stmt: StatementNode = self.dg.nodes[sid]["data"]
                stmt_outedges = self.out_edges(sid)
                if len(stmt_outedges) > 1:
                    # this stmt has other qualifiers, so it's not what we are looking for
                    continue

                for sv_outedge in next(iter(stmt_outedges.values())):
                    v = self.dg.nodes[sv_outedge.target]["data"]
                    # got leg 2, now looks for all incoming
                    leg2 = (
                        EdgeFlowSource(nid, ns_eid),
                        EdgeFlowTarget(v.id, sv_outedge.predicate),
                    )
                    if not stmt.has_flow(*leg2):
                        continue
                    leg2_provenance = stmt.get_provenance(*leg2)

                    has_better_paths = True
                    for gpid in grandpa:
                        if (gpid, v.id) not in legprime:
                            paths = [
                                (
                                    self.dg.nodes[path[0][1]]["data"],
                                    EdgeFlowSource(path[0][0], path[0][2]),
                                    EdgeFlowTarget(path[1][1], path[1][2]),
                                )
                                for path in nx.all_simple_edge_paths(
                                    self.dg, gpid, v.id, cutoff=2
                                )
                            ]
                            provs = [
                                prov
                                for s, sf, tf in paths
                                if s.has_flow(sf, tf)
                                for prov in s.get_provenance(sf, tf)
                            ]
                            if len(provs) == 0:
                                legprime[gpid, v.id] = None
                            else:
                                legprime[gpid, v.id] = max(
                                    provs,
                                    key=cmp_to_key(
                                        self.specific_pruning_provenance_cmp
                                    ),
                                )
                        best_prov = legprime[gpid, v.id]
                        if (
                            best_prov is None
                            or max(
                                self.specific_pruning_provenance_cmp(
                                    best_prov, leg2_prov
                                )
                                for leg2_prov in leg2_provenance
                            )
                            < 0
                        ):
                            # no better path
                            has_better_paths = False
                            break

                    if has_better_paths:
                        rm_legs.append((sid, leg2[0], leg2[1]))

        # logger.info("#legs: {}", len(rm_legs))
        self.remove_flow(rm_legs)
        # logger.info("# 0-indegree: {}", sum(self.dg.in_degree(uid) == 0 for uid in self.dg.nodes))
        # logger.info("# 0-outdegree: {}", sum(self.dg.out_degree(uid) == 0 for uid in self.dg.nodes))
        # logger.info("# 0-standalone: {}",
        #             sum(self.dg.out_degree(uid) + self.dg.in_degree(uid) == 0 for uid in self.dg.nodes))

        # step 2: prune the first leg paths
        rm_legs: List[Tuple[str, EdgeFlowSource, EdgeFlowTarget]] = []
        for nid, ndata in self.dg.nodes(data=True):
            n: EntityValueNode = ndata["data"]
            if not n.is_entity_value or self.dg.out_degree(nid) > 0:
                continue

            for sid, _, sn_eid, sn_edata in self.dg.in_edges(nid, data=True, keys=True):
                stmt_outedges = self.out_edges(sid)
                if len(stmt_outedges) == 1:
                    # stmt does not have other property/qualifier
                    target_flow = EdgeFlowTarget(nid, sn_eid)
                    stmt: StatementNode = self.dg.nodes[sid]["data"]
                    for source_flow, _ in stmt.iter_source_flow(target_flow):
                        rm_legs.append((sid, source_flow, target_flow))

        # logger.info("#legs: {}", len(rm_legs))
        self.remove_flow(rm_legs)
        # logger.info("# 0-indegree: {}", sum(self.dg.in_degree(uid) == 0 for uid in self.dg.nodes))
        # logger.info("# 0-outdegree: {}", sum(self.dg.out_degree(uid) == 0 for uid in self.dg.nodes))
        # logger.info("# 0-standalone: {}",
        #             sum(self.dg.out_degree(uid) + self.dg.in_degree(uid) == 0 for uid in self.dg.nodes))
        self.prune_disconnected_nodes()

    def prune_disconnected_nodes(self):
        """This function prune out disconnected nodes that are:
        1. nodes without incoming edges and outgoing edges
        2. statement nodes with no incoming edges or no outgoing edges

        Returns
        -------
        """
        rm_nodes = set()
        for uid, udata in self.dg.nodes(data=True):
            u: DGNode = udata["data"]
            if u.is_entity_value:
                if self.dg.in_degree(uid) == 0 and self.dg.out_degree(uid) == 0:
                    rm_nodes.add(uid)
            elif u.is_statement:
                if self.dg.in_degree(uid) == 0 or self.dg.out_degree(uid) == 0:
                    rm_nodes.add(uid)
        for uid in rm_nodes:
            self.dg.remove_node(uid)

    def remove_flow(self, flows: List[Tuple[str, EdgeFlowSource, EdgeFlowTarget]]):
        for sid, source_flow, target_flow in flows:
            stmt: StatementNode = self.dg.nodes[sid]["data"]
            stmt.untrack_flow(source_flow, target_flow)
            if not stmt.has_source_flow(source_flow):
                self.dg.remove_edge(source_flow.source_id, sid, source_flow.edge_id)
            if not stmt.has_target_flow(target_flow):
                self.dg.remove_edge(sid, target_flow.target_id, target_flow.edge_id)

    def specific_pruning_provenance_cmp(
        self, prov0: FlowProvenance, prov1: FlowProvenance
    ):
        # compare provenance, this function only accept
        if prov0.gen_method == LinkGenMethod.FromWikidataLink:
            # always favour from wikidata link
            return 1
        if prov1.gen_method == LinkGenMethod.FromWikidataLink:
            return -1
        # assert prov0.gen_method == prov1.gen_method and prov0.gen_method_arg == prov1.gen_method_arg
        # do not need to check if the two gen method and args are equal, as even if we select the incorrect one
        # we only truncate when the other leg worst than it
        return prov0.prob - prov1.prob

    def iter_grand_parents(self, nid: str):
        for pid, _ in self.dg.in_edges(nid):
            for ppid, _ in self.dg.in_edges(pid):
                yield ppid

    def out_edges(self, uid: str) -> Dict[str, List[DGEdge]]:
        label2edges = defaultdict(list)
        for _, vid, eid, edata in self.dg.out_edges(uid, data=True, keys=True):
            label2edges[eid].append(edata["data"])
        return label2edges


def prune_data_graph_deprecated(dg: nx.MultiDiGraph):
    """This algorithm prunes:

    * Redundant hidden entities added to the graph from KG discovering. This can also include entities in the context.


    0. A second leg of a path between two nodes that goes through an entity node f1:
        (f0 -> p1 -> s1 -> p2 -> f1) (-> p3 -> s2 -> p3 -> f3). This removal is triggered when
        s2 doesn't have other qualifiers map to other nodes
    there is another better path between f1 and f3 (shorter, also from wikidatalink), s2 must not have any

    1. An entity that between all sources and targets that it connects exist a better path in the graph connecting them,
    and it is not used as a mean (should we consider it to be the only mean?) to connect the other columns/context as property & qualifiers (i.e., property value of statement). for example,
    an award is needed to connect the year column in the model. Note: A path is consider better when it is shorter,
    and also generated from WikidataLink rather than literal matching.
    2. Statements that belong to the same qnode and same property, have the same value and qualifiers matched in the data graph.
    The reason we may have multiple statements like this, is because a property may have multiple statements of the same values in
    KG, usually different with each other by the qualifiers

    Finally, after applying those pruning rules, if any dangling nodes will be removed too (e.g., a statement node that
    doesn't have any value, or an entity node that doesn't link to any node)

    Parameters
    ----------
    dg

    Returns
    -------
    """

    def iter_grand_parents(edges: List[DGPruning.NxDGEdge]):
        for f2, f3, eid, edge in edges:
            for f1, _f2 in dg.in_edges(f2):
                yield dg.nodes[f1]["data"]

    def iter_grand_children(edges: List[DGPruning.NxDGEdge]):
        """Iterate through children of children of a given node"""
        for f1, f2, eid, edge in edges:
            for f3, f23_edges in dg[f2].items():
                yield dg.nodes[f3]["data"]

    def iter_sibling_edges(edges: List[DGPruning.NxDGEdge]):
        """Iter through out sibling nodes"""
        for f1, f2, eid, edge in edges:
            for f2_sib, sib_edges in dg[f1].items():
                for sib_eid, sib_edata in sib_edges.items():
                    yield sib_edata["data"]

    def remove_nodes(node_ids: Set[str], remove_dangling_statements: bool = True):
        for nid in node_ids:
            node = dg.nodes[nid]["data"]
            if node.is_statement:
                # if it is statements, just remove it
                dg.remove_node(nid)
            else:
                # when delete a node, we need to remove its flow also
                for sid, _, eid in dg.in_edges(nid, keys=True):
                    s: StatementNode = dg.nodes[sid]["data"]
                    s.untrack_target_flow(EdgeFlowTarget(nid, eid))
                for _, sid, eid in dg.out_edges(nid, keys=True):
                    s: StatementNode = dg.nodes[sid]["data"]
                    s.untrack_source_flow(EdgeFlowSource(nid, eid))
                dg.remove_node(nid)

        if not remove_dangling_statements or len(node_ids) == 0:
            return

        # remove dangling statements
        dangling_statements = set()
        for nid in dg.nodes:
            node = dg.nodes[nid]["data"]
            if node.is_statement and (
                dg.out_degree(nid) == 0 or dg.in_degree(nid) == 0
            ):
                dangling_statements.add(nid)
        for nid in dangling_statements:
            dg.remove_node(nid)

    def extract_stmt_signature(stmt_id: str):
        # to compare if two statements are equal, we only need to compare if the property & qualifiers values are
        # the same. We define that the value is the same by their node id in the graph (so that cell need to be the exact
        # cell, while literal or entity will be the value)
        signature = set()
        for uid, edges in dg[stmt_id].items():
            for eid in edges:
                signature.add(f"({eid}-{uid})")
        signature = sorted(list(signature))
        return "-".join(signature)

    has_direct_kglink_paths = {}

    rm_node_ids = set()
    # remove entities first.
    for uid, udata in dg.nodes(data=True):
        u: EntityValueNode = udata["data"]
        if not u.is_entity_value:
            continue

        # find entity node that only connected by two distinct qnodes
        inedges: List[DGPruning.NxDGEdge] = list(dg.in_edges(uid, data=True, keys=True))
        outedges: List[DGPruning.NxDGEdge] = list(
            dg.out_edges(uid, data=True, keys=True)
        )

        if len(inedges) == 0:
            # this happen when this entity is needed to links between columns or column and context as property & qualifiers
            # we can't prune this node, but we want to make sure this is actually the case
            assert (
                len(
                    {
                        v.column if v.is_cell else v.id
                        for v in iter_grand_children(outedges)
                        if v.is_cell
                        or ((v.is_literal_value or v.is_entity_value) and v.is_context)
                    }
                )
                >= 2
            ), "Should have at least two columns or one column one context"
            continue

        if len(outedges) == 0:
            # this happen when the real value of statement is missing, we can't prune this node
            # but we want to make sure there is actually another qualifiers are there
            assert (
                len({sib_edge.predicate for sib_edge in iter_sibling_edges(inedges)})
                >= 2
            ), "The statement should have at least two distinct predicates"
            continue

        # check if it is used to connect other columns via property & qualifiers

        # check if this entity only connect two qnodes and no literal values
        source_qnodes = {dg.nodes[sid]["data"].qnode_id for _1, sid, _2, _3 in inedges}
        if len(source_qnodes) != 1:
            continue
        # TODO: in order to get correct target qnodes, we need to examine the flow of the statements, the current finding
        # target qnodes will not work as expected when the target cell have more than one qnode
        target_qnodes = set()
        for v in iter_grand_children(outedges):
            if v.is_cell:
                target_qnodes = target_qnodes.union(v.qnode_ids)
            elif v.is_entity_value:
                target_qnodes.add(v.qnode_id)
            else:
                assert v.is_literal_value
                target_qnodes = set()
                break
        if len(target_qnodes) != 1:
            continue

        # now they connect only two distinct qnodes and no literal, so we know this path is through wikidata links
        # find out if all grand parents and grand children have better paths
        # first retrieve the list of grant parent and grant child we need to test.
        # instead of finding pairs of parents and children through flows, we do it here..?
        grant_parents = list(iter_grand_parents(inedges))
        grand_children = list(iter_grand_children(outedges))

        grand_child_cells = defaultdict(list)
        grand_child_ents = []
        for gc in grand_children:
            if gc.is_cell:
                grand_child_cells[gc.row].append(gc)
            else:
                assert gc.is_entity_value
                grand_child_ents.append(gc)

        all_paths_are_better = True
        for gp in grant_parents:
            if gp.is_cell:
                corresponding_gcs = chain(grand_child_cells[gp.row], grand_child_ents)
            else:
                corresponding_gcs = grand_child_ents
            for gc in corresponding_gcs:
                # identify if there is other better path is shorter and directly connect the two nodes
                if (gp.id, gc.id) not in has_direct_kglink_paths:
                    has_better_path = False
                    has_better_path = any(
                        (
                            any(
                                (
                                    provenance.gen_method
                                    == LinkGenMethod.FromWikidataLink
                                    for provenance in dg.nodes[path[0][1]][
                                        "data"
                                    ].get_provenance(
                                        EdgeFlowSource(path[0][0], path[0][2]),
                                        EdgeFlowTarget(path[1][1], path[1][2]),
                                    )
                                )
                            )
                            for path in nx.all_simple_edge_paths(
                                dg, gp.id, gc.id, cutoff=2
                            )
                        )
                    )
                    has_direct_kglink_paths[gp.id, gc.id] = has_better_path
                else:
                    has_better_path = has_direct_kglink_paths[gp.id, gc.id]
                if not has_better_path:
                    all_paths_are_better = False
                    break
            if not all_paths_are_better:
                break

        if all_paths_are_better:
            rm_node_ids.add(uid)

    remove_nodes(rm_node_ids)

    rm_node_ids = set()
    # remove duplicated statements
    for uid, udata in dg.nodes(data=True):
        u: Union[CellNode, EntityValueNode] = udata["data"]
        if not (u.is_cell or u.is_entity_value):
            continue

        # we are looking at statements of the same edges
        prop2stmts = defaultdict(list)
        for sid, us_edges in dg[uid].items():
            for eid, edata in us_edges.items():
                prop2stmts[eid].append(sid)

        for stmts in prop2stmts.values():
            # remove duplicated statements
            if len(stmts) > 1:
                stmt_sig = {}
                for si, sid in enumerate(stmts):
                    sig = extract_stmt_signature(sid)
                    if sig in stmt_sig:
                        continue
                    stmt_sig[sig] = si
                if len(stmt_sig) < len(stmts):
                    for si in set(range(len(stmts))).difference(stmt_sig.values()):
                        rm_node_ids.add(stmts[si])

    remove_nodes(rm_node_ids)


def build_data_graph(
    table: LinkedTable,
    qnodes: Dict[str, QNode],
    wdprops: Dict[str, WDProperty],
    kg_object_index: KGObjectIndex,
    ignore_columns: Set[int] = None,
    max_n_hop: int = 2,
    options: BuildDGOption = BuildDGOption.DISABLE_SAME_ENT_SEARCH
    | BuildDGOption.PRUNING_REDUNDANT_ENT,
    verbose: bool = False,
):
    """Build data graph from a linked table"""
    dg = nx.MultiDiGraph()
    context_node_id = None

    for ci, col in enumerate(table.table.columns):
        for ri, val in enumerate(col.values):
            # =====================================================
            # NOTE: old code
            # cell_qnodes = set()
            # cell_qnode_spans = {}
            # for link in table.links[ri][ci]:
            #     if link.entity_id is not None:
            #         cell_qnodes.add(link.entity_id)
            #         if link.entity_id not in cell_qnode_spans:
            #             cell_qnode_spans[link.entity_id] = []
            #         cell_qnode_spans[link.entity_id].append(Span(link.start, link.end))
            # assert all(len(spans) == len(set(spans)) for spans in cell_qnode_spans.values())
            # TODO: new code, doesn't handle the qnode_spans correctly
            cell_qnodes = {
                candidate.entity_id
                for link in table.links[ri][ci]
                for candidate in link.candidates
            }
            cell_qnode_spans = {}
            for link in table.links[ri][ci]:
                if link.entity_id is not None or len(link.candidates) > 0:
                    if len(link.candidates) > 0:
                        tmpid = link.candidates[0].entity_id
                    else:
                        tmpid = link.entity_id
                    if tmpid not in cell_qnode_spans:
                        cell_qnode_spans[tmpid] = []
                    cell_qnode_spans[tmpid].append(Span(link.start, link.end))
            assert all(
                len(spans) == len(set(spans)) for spans in cell_qnode_spans.values()
            )
            # =====================================================

            node = CellNode(
                id=f"{ri}-{ci}",
                value=val,
                column=ci,
                row=ri,
                qnode_ids=list(cell_qnodes),
                qnodes_span=cell_qnode_spans,
            )
            dg.add_node(node.id, data=node)

    if table.context.page_entity_id is not None:
        context_node_id = DGPathNodeQNode(table.context.page_entity_id).get_id()
        node = EntityValueNode(
            id=context_node_id,
            qnode_id=table.context.page_entity_id,
            context_span=ContextSpan(
                text=table.context.page_title,
                span=Span(0, len(table.context.page_title)),
            ),
        )
        dg.add_node(node.id, data=node)

    # find all paths
    n_rows = len(table.table.columns[0].values)
    kg_path_discovering_tasks = []
    new_paths = []
    for ri in range(n_rows):
        for ci, col in enumerate(table.table.columns):
            if ignore_columns is not None and ci in ignore_columns:
                continue

            u = dg.nodes[f"{ri}-{ci}"]["data"]
            for cj in range(ci + 1, len(table.table.columns)):
                if ignore_columns is not None and cj in ignore_columns:
                    continue

                v = dg.nodes[f"{ri}-{cj}"]["data"]
                kg_path_discovering_tasks.append((u, v))
            if context_node_id in dg:
                kg_path_discovering_tasks.append((u, dg.nodes[context_node_id]["data"]))

    for u, v in (
        tqdm(kg_path_discovering_tasks, desc="KG searching")
        if verbose
        else kg_path_discovering_tasks
    ):
        new_paths += kg_path_discovering(
            options, qnodes, kg_object_index, dg, u, v, max_n_hop=max_n_hop
        )

    # add paths to the graph
    for path in new_paths:
        curr_nodeid = path.sequence[0].id
        tmp_path = [curr_nodeid]
        for i in range(1, len(path.sequence), 2):
            prop: DGPathEdge = path.sequence[i]
            value: DGPathNode = path.sequence[i + 1]

            if isinstance(value, DGPathExistingNode):
                nodeid = value.id
            else:
                nodeid = value.get_id()
                if nodeid not in dg.nodes:
                    if isinstance(value, DGPathNodeStatement):
                        dg.add_node(
                            nodeid,
                            data=StatementNode(
                                nodeid,
                                qnode_id=value.qnode_id,
                                predicate=value.predicate,
                                is_in_kg=True,
                            ),
                        )
                    elif isinstance(value, DGPathNodeQNode):
                        dg.add_node(
                            nodeid,
                            data=EntityValueNode(
                                nodeid, qnode_id=value.qnode_id, context_span=None
                            ),
                        )
                    else:
                        dg.add_node(
                            nodeid,
                            data=LiteralValueNode(
                                nodeid, value=value.value, context_span=None
                            ),
                        )

            if dg.has_edge(curr_nodeid, nodeid, prop.value):
                edge: DGEdge = dg.edges[curr_nodeid, nodeid, prop.value]["data"]
            else:
                edge = DGEdge(
                    source=curr_nodeid,
                    target=nodeid,
                    predicate=prop.value,
                    is_qualifier=prop.is_qualifier,
                )
                dg.add_edge(curr_nodeid, nodeid, key=prop.value, data=edge)

            tmp_path.append(edge)
            tmp_path.append(nodeid)
            curr_nodeid = nodeid

        # path format: u - e - (s) - e - v - e - (s2) - e - v2
        for i in range(2, len(tmp_path), 4):
            u_node_id = tmp_path[i - 2]
            u_edge = tmp_path[i - 1]
            snode: StatementNode = dg.nodes[tmp_path[i]]["data"]
            v_edge = tmp_path[i + 1]
            v_node_id = tmp_path[i + 2]

            edge_source = EdgeFlowSource(u_node_id, u_edge.predicate)
            edge_target = EdgeFlowTarget(v_node_id, v_edge.predicate)
            snode.track_provenance(
                edge_source, edge_target, [path.sequence[i].provenance]
            )

    # DEBUG code
    # for uid, udata in dg.nodes(data=True):
    #     u = udata['data']
    #     if not u.is_cell:
    #         continue
    #     if u.column == 0:
    #         for _, sid, eid, edata in dg.out_edges(uid, data=True, keys=True):
    #             for _1, vid, e2, e2data in dg.out_edges(sid, data=True, keys=True):
    #                 v = dg.nodes[vid]['data']
    #                 if v.is_cell and v.column == 1:
    #                     print(uid, sid, vid, eid, e2)

    KGInference(dg, qnodes, wdprops).infer_subproperty().kg_transitive_inference()

    # pruning unnecessary paths
    if options & BuildDGOption.PRUNING_REDUNDANT_ENT:
        DGPruning(dg).prune_hidden_entities()

    M.log("grams", data_graph=dg)
    return dg



from typing import List, Dict, Tuple, Callable, Any, Optional

import grams.misc as M
from grams.config import DATA_DIR
from grams.kg_data.wikidatamodels import WDProperty, QNode
from grams.main import GRAMS


def update_props(props: List[str]):
    props: Dict[str, QNode] = GRAMS._query_wikidata_entities(props)
    ser = []
    for p in props.values():
        np = WDProperty(
            p.id,
            str(p.label),
            str(p.description),
            str(p.datatype),
            [str(s) for s in p.aliases],
            sorted({stmt.value.as_entity_id() for stmt in p.props.get("P1647", [])}),
            sorted({stmt.value.as_entity_id() for stmt in p.props.get("P1659", [])}),
            sorted({stmt.value.as_string() for stmt in p.props.get("P1628", [])}),
            sorted({stmt.value.as_entity_id() for stmt in p.props.get("P1629", [])}),
            sorted({stmt.value.as_entity_id() for stmt in p.props.get("P1696", [])}),
            sorted({stmt.value.as_entity_id() for stmt in p.props.get("P31", [])}),
        )
        ser.append(np.serialize())
    M.serialize_byte_lines(ser, DATA_DIR / "new_properties.jl")


if __name__ == "__main__":
    update_props(["P8901"])



import copy
from dataclasses import dataclass
from operator import itemgetter
from urllib.parse import urlparse, unquote_plus

import requests, re, pandas as pd
from bs4 import BeautifulSoup, NavigableString, Tag
from typing import List, Dict, Optional

from sm.prelude import I, M
import grams.inputs as GI
from grams.config import DATA_DIR
from grams.db import Wikipedia2WikidataDB


@dataclass
class RemoteExtractedTableCellHTMLElement:
    # html tag (lower case)
    tag: str
    start: int
    # end (exclusive)
    end: int
    # html attributes
    attrs: Dict[str, str]
    children: List["RemoteExtractedTableCellHTMLElement"]

    def post_order(self):
        for c in self.children:
            for el in c.post_order():
                yield el
        yield self

    def clone(self):
        return RemoteExtractedTableCellHTMLElement(
            self.tag,
            self.start,
            self.end,
            copy.copy(self.attrs),
            [c.clone() for c in self.children],
        )


@dataclass
class RemoteExtractedTableCell:
    # text value of the cell
    value: str
    rowspan: int
    colspan: int
    # html of the cell
    html: str
    # list of html elements that created this cell
    # except that:
    # - BR and HR are replaced by `\n` character and not in this list
    # - div, span are container and won't in the list
    # for more details, look at the _extract_cell_recur function
    elements: List[RemoteExtractedTableCellHTMLElement]

    # original row & col span
    original_rowspan: Optional[int] = None
    original_colspan: Optional[int] = None

    def travel_elements_post_order(self):
        for el in self.elements:
            for pointer in el.post_order():
                yield pointer

    def clone(self):
        return RemoteExtractedTableCell(
            value=self.value,
            rowspan=self.rowspan,
            colspan=self.colspan,
            html=self.html,
            elements=[el.clone() for el in self.elements],
            original_rowspan=self.original_rowspan,
            original_colspan=self.original_colspan,
        )


@dataclass
class RemoteExtractedTableRow:
    cells: List[RemoteExtractedTableCell]
    # html attributes of the tr elements
    attrs: Dict[str, str]


@dataclass
class ContextLevel:
    level: int
    header: str
    content: str

    def clone(self):
        return ContextLevel(self.level, self.header, self.content)


class OverlapSpanException(Exception):
    """Indicating the table has cell rowspan and cell colspan overlaps"""

    pass


class InvalidColumnSpanException(Exception):
    """Indicating that the column span is not used in a standard way. In particular, the total of columns' span is beyond the maximum number of columns is considered
    to be non standard with one exception that only the last column spans more than the maximum number of columns
    """

    pass


@dataclass
class RemoteExtractedTable:
    page_url: str
    # value of html caption
    caption: str
    # html attributes of the table html element
    attrs: Dict[str, str]
    # context
    context: List[ContextLevel]
    # list of rows in the table
    rows: List[RemoteExtractedTableRow]

    def span(self) -> "RemoteExtractedTable":
        """Span the table by copying values to merged field"""
        pi = 0
        data = []
        pending_ops = {}

        # >>> begin find the max #cols
        # calculate the number of columns as some people may actually set unrealistic colspan as they are lazy..
        # I try to make its behaviour as much closer to the browser as possible.
        # one thing I notice that to find the correct value of colspan, they takes into account the #cells of rows below the current row
        # so we may have to iterate several times
        cols = [0 for _ in range(len(self.rows))]
        for i, row in enumerate(self.rows):
            cols[i] += len(row.cells)
            for cell in row.cells:
                if cell.rowspan > 1:
                    for j in range(1, cell.rowspan):
                        if i + j < len(cols):
                            cols[i + j] += 1

        _row_index, max_ncols = max(enumerate(cols), key=itemgetter(1))
        # sometimes they do show an extra cell for over-colspan row, but it's not consistent or at least not easy for me to find the rule
        # so I decide to not handle that. Hope that we don't have many tables like that.
        # >>> finish find the max #cols

        for row in self.rows:
            new_row = []
            pj = 0
            for cell_index, cell in enumerate(row.cells):
                cell = cell.clone()
                cell.original_colspan = cell.colspan
                cell.original_rowspan = cell.rowspan
                cell.colspan = 1
                cell.rowspan = 1

                # adding cell from the top
                while (pi, pj) in pending_ops:
                    new_row.append(pending_ops[pi, pj].clone())
                    pending_ops.pop((pi, pj))
                    pj += 1

                # now add cell and expand the column
                for _ in range(cell.original_colspan):
                    if (pi, pj) in pending_ops:
                        # exception, overlapping between colspan and rowspan
                        raise OverlapSpanException()
                    new_row.append(cell.clone())
                    for ioffset in range(1, cell.original_rowspan):
                        # no need for this if below
                        # if (pi+ioffset, pj) in pending_ops:
                        #     raise OverlapSpanException()
                        pending_ops[pi + ioffset, pj] = cell
                    pj += 1

                    if pj >= max_ncols:
                        # our algorithm cannot handle the case where people are bullying the colspan system, and only can handle the case
                        # where the span that goes beyond the maximum number of columns is in the last column.
                        if cell_index != len(row.cells) - 1:
                            raise InvalidColumnSpanException()
                        else:
                            break

            # add more cells from the top since we reach the end
            while (pi, pj) in pending_ops and pj < max_ncols:
                new_row.append(pending_ops[pi, pj].clone())
                pending_ops.pop((pi, pj))
                pj += 1

            data.append(
                RemoteExtractedTableRow(cells=new_row, attrs=copy.copy(row.attrs))
            )
            pi += 1

        # len(pending_ops) may > 0, but fortunately, it doesn't matter as the browser also does not render that extra empty lines
        return RemoteExtractedTable(
            page_url=self.page_url,
            caption=self.caption,
            attrs=copy.copy(self.attrs),
            context=[c.clone() for c in self.context],
            rows=data,
        )

    def pad(self) -> "RemoteExtractedTable":
        """Pad the irregular table (missing cells) to make it become regular table.

        This function only return new table when it's padded
        """
        if len(self.rows) == 0:
            return self

        ncols = len(self.rows[0].cells)
        is_regular_table = all(len(r.cells) == ncols for r in self.rows)
        if is_regular_table:
            return self

        max_ncols = max(len(r.cells) for r in self.rows)
        default_cell = RemoteExtractedTableCell(
            value="",
            rowspan=1,
            colspan=1,
            html="",
            elements=[],
            original_rowspan=1,
            original_colspan=1,
        )

        rows = []
        for r in self.rows:
            row = RemoteExtractedTableRow(
                cells=[c.clone() for c in r.cells], attrs=copy.copy(r.attrs)
            )
            while len(row.cells) < max_ncols:
                row.cells.append(default_cell.clone())
            rows.append(row)

        return RemoteExtractedTable(
            page_url=self.page_url,
            caption=self.caption,
            attrs=copy.copy(self.attrs),
            context=[c.clone() for c in self.context],
            rows=rows,
        )

    def as_df(self):
        return pd.DataFrame([[c.value for c in r.cells] for r in self.rows])

    def as_relational_linked_table(self, table_id=None):
        assert len(self.rows) > 0
        header = [c.value for c in self.rows[0].cells]
        table = I.ColumnBasedTable(
            table_id or self.page_url,
            [
                I.Column(
                    ci,
                    cname,
                    [self.rows[ri].cells[ci].value for ri in range(1, len(self.rows))],
                )
                for ci, cname in enumerate(header)
            ],
        )

        wikidb = Wikipedia2WikidataDB.get_instance(
            DATA_DIR / "enwiki_links.db", read_only=True
        )
        links = []
        for ri, row in enumerate(self.rows[1:]):
            rlinks = []
            for ci in range(len(header)):
                clinks = []
                for el in row.cells[ci].travel_elements_post_order():
                    if el.tag == "a":
                        entity_id = wikidb.get(
                            self.get_title_from_url(el.attrs["href"]), None
                        )
                        if entity_id is not None:
                            candidates = [GI.CandidateEntity(entity_id, 1.0)]
                        else:
                            candidates = []
                        clinks.append(
                            GI.Link(
                                el.start,
                                el.end,
                                el.attrs["href"],
                                entity_id,
                                candidates,
                            )
                        )
                rlinks.append(clinks)
            links.append(rlinks)
        return GI.LinkedTable(table, GI.Context(), links)

    def get_title_from_url(self, url: str) -> str:
        """This function converts a wikipedia page/article's URL to its title. The function is tested manually in `20200425-wikipedia-links` notebook in section 2.2.

        Parameters
        ----------
        url : str
            a wikipedia page/article's URL

        Returns
        -------
        str
            a wikipedia page/article's title
        """
        path = urlparse(url).path
        if not path.startswith("/wiki/"):
            return ""

        assert path.startswith("/wiki/"), path
        path = path[6:]
        title = unquote_plus(path).replace("_", " ")
        return title.strip()


class InvalidCellSpanException(Exception):
    """Indicating that the html colspan or rowspan is wrong"""

    pass


class HTMLTableExtractor:
    NUM_REGEX = re.compile("(\d+)")
    IGNORE_TAGS = set(["div", "span"])

    @dataclass
    class Header:
        level: int
        value: str

    @dataclass
    class Text:
        value: str

    """Extract tables from an html table, which may contains multiple nested tables. I assume that the outer tables are used for
    formatting and the inner tables are the interested ones. So this extractor will discard the outer tables and only keep
    tables at the bottom level.

    The main function of this extractor are extract
    """

    def extract(self, page_url: str, html: str) -> List[RemoteExtractedTable]:
        """Extract just one html table, which may contains multiple nested tables. I assume that the outer tables are used for
        formatting and the inner tables are the interested ones. So this function will discard the outer tables and only keep
        tables at the bottom level.

        This function right now ignore tables with invalid rowspan or cellspan
        """
        el = BeautifulSoup(html, "html5lib")
        table_els = el.find_all("table")
        results = []
        for table_el in table_els:
            temp_results = []
            self._extract_table(table_el, temp_results)
            if len(temp_results) > 0:
                context = self._locate_context(table_el)
                for r in temp_results:
                    r["context"] = context
                results += temp_results

        tables = [RemoteExtractedTable(page_url, **r) for r in results]

        # convert relative links to absolute links
        parsed_resp = urlparse(page_url)
        domain = f"{parsed_resp.scheme}://{parsed_resp.netloc}"
        for table in tables:
            for row in table.rows:
                for cell in row.cells:
                    for el in cell.travel_elements_post_order():
                        if el.tag == "a":
                            if "href" in el.attrs:
                                href = el.attrs["href"]
                                if href[0] == "/":
                                    el.attrs["href"] = domain + href

        if parsed_resp.netloc.endswith("wikipedia.org"):
            for table in tables:
                self._postprocess_wikipedia(table)

        return tables

    def _extract_table(self, el: Tag, results: List[dict]):
        """Extract tables from the table tag. Ignore tables that contain other tables

        Parameters
        ----------
        el : Tag
            html table tag
        results : List[dict]
            list of results
        """
        assert el.name == "table"
        caption = None
        rows = []
        contain_nested_table = any(
            c.find("table") is not None is not None for c in el.contents
        )

        for c in el.contents:
            if isinstance(c, NavigableString):
                continue
            if c.name == "caption":
                caption = c.get_text().strip()
                continue
            if c.name == "style":
                continue
            assert (
                c.name == "thead" or c.name == "tbody"
            ), f"not implemented {c.name} tag"
            for row_el in c.contents:
                if isinstance(row_el, NavigableString):
                    continue
                if row_el.name == "style":
                    continue

                assert row_el.name == "tr", f"Invalid tag: {row_el.name}"
                cells = []
                for cell_el in row_el.contents:
                    if isinstance(cell_el, NavigableString):
                        continue
                    if cell_el.name == "style":
                        continue

                    assert (
                        cell_el.name == "th" or cell_el.name == "td"
                    ), f"Invalid tag: {row_el.name}"

                    if contain_nested_table:
                        nested_tables = []
                        self._extract_tbl_tags(cell_el, nested_tables)
                        for tag in nested_tables:
                            self._extract_table(tag, results)
                    else:
                        try:
                            cell = self._extract_cell(cell_el)
                            cells.append(cell)
                        except InvalidCellSpanException:
                            # not extracting this table, this table is not table at the intermediate level
                            return

                if not contain_nested_table:
                    rows.append(
                        RemoteExtractedTableRow(cells=cells, attrs=row_el.attrs)
                    )

        if not contain_nested_table:
            results.append(dict(rows=rows, caption=caption, attrs=el.attrs))

    def _extract_tbl_tags(self, el: Tag, tags: List[Tag]):
        """Recursive find the first table node

        Parameters
        ----------
        el : Tag
            root node that we start finding
        tags : List[Tag]
            list of results
        """
        for c in el.children:
            if c.name == "table":
                tags.append(c)
            elif isinstance(c, NavigableString):
                continue
            elif len(c.contents) > 0:
                self._extract_tbl_tags(c, tags)

    def _extract_cell(self, el: Tag) -> RemoteExtractedTableCell:
        """Extract cell from td/th tag. This function does not expect a nested table in the cell

        Parameters
        ----------
        tbl : RawHTMLTable
            need the table to get the page URI in case we have cell reference
        el : Tag
            cell tag (th or td)
        """
        assert el.name == "th" or el.name == "td"

        # extract other attributes
        is_header = el.name == "th"
        attrs = el.attrs

        colspan = attrs.get("colspan", "1").strip()
        rowspan = attrs.get("rowspan", "1").strip()

        if colspan == "":
            colspan = 1
        else:
            m = self.NUM_REGEX.search(colspan)
            if m is None:
                raise InvalidCellSpanException()
            colspan = int(m.group(0))
        if rowspan == "":
            rowspan = 1
        else:
            m = self.NUM_REGEX.search(rowspan)
            if m is None:
                # this is not correct, but
                raise InvalidCellSpanException()
            rowspan = int(m.group(0))

        # extract value
        result = {"text": "", "elements": []}
        self._extract_cell_recur(el, result)
        value, elements = result["text"], result["elements"]
        if len(value) > 0 and value[-1] == " ":
            value = value[:-1]

        return RemoteExtractedTableCell(
            value=value,
            html=str(el),
            elements=elements,
            colspan=colspan,
            rowspan=rowspan,
        )

    def _extract_cell_recur(self, el: Tag, result: dict):
        """Real implementation of the _extract_cell function"""
        for c in el.children:
            c_name = c.name
            if c_name is None:
                result["text"] += c.strip()
                result["text"] += " "
            elif c_name == "br" or c_name == "hr":
                result["text"] += "\n"
            elif c_name in self.IGNORE_TAGS:
                self._extract_cell_recur(c, result)
            else:
                start = len(result["text"])
                children_result = {"text": result["text"], "elements": []}
                self._extract_cell_recur(c, children_result)
                # -1 for the trailing space
                if (
                    len(children_result["text"]) > 0
                    and children_result["text"][-1] == " "
                ):
                    if start == len(children_result["text"]):
                        # no new text
                        start -= 1
                    end = len(children_result["text"]) - 1
                else:
                    end = len(children_result["text"])
                assert end >= start
                result["text"] = children_result["text"]
                result["elements"].append(
                    RemoteExtractedTableCellHTMLElement(
                        tag=c_name,
                        start=start,
                        end=end,
                        attrs=c.attrs,
                        children=children_result["elements"],
                    )
                )

    def _locate_context_flatten_hierarchy(self, el):
        if el.name is None:
            # navigable string
            return [self.Text(str(el))]

        # flatten the tree to list of text? -> (header -> text)*
        if el.name in {"h1", "h2", "h3", "h4", "h5", "h6"}:
            return [self.Header(level=int(el.name[1:]), value=el.get_text())]
        else:
            assert el.name == "hr" or not el.name.startswith("h")
            lst = []
            for c in el.contents:
                lst += self._locate_context_flatten_hierarchy(c)
            return lst

    def _locate_context_optimize_flatten_hierarchy(self, lst):
        if len(lst) == 0:
            return lst
        new_lst = []
        for item in lst:
            if isinstance(item, self.Header):
                new_lst.append(item)
            else:
                if len(new_lst) > 0 and isinstance(new_lst[-1], self.Text):
                    new_lst[-1].value += "\n" + item.value
                else:
                    new_lst.append(item)
        return new_lst

    def _locate_context_recur(self, el):
        if el.name in {"body", "html"}:
            # hit the top
            return []

        parent = el.parent
        prev_sibs = []
        for i, e in enumerate(parent.contents):
            if e == el:
                # this is the index
                break
            prev_sibs.append(e)

        lst = []
        for sib in prev_sibs:
            lst += self._locate_context_flatten_hierarchy(sib)
        lst = self._locate_context_optimize_flatten_hierarchy(lst)
        return self._locate_context_optimize_flatten_hierarchy(
            self._locate_context_recur(el.parent) + lst
        )

    def _locate_context(self, el):
        """Locate the context of the elements"""
        context = self._locate_context_recur(el)
        # optimize the context
        optimized_context = []
        last_header = None
        if isinstance(context[-1], self.Header):
            last_header = ContextLevel(
                context[-1].level, header=context[-1].value, content=""
            )
            context.pop()
        if isinstance(context[0], self.Text):
            assert isinstance(context[1], self.Header) and context[1].level == 1
            context.pop(0)
        else:
            assert isinstance(context[0], self.Header) and context[0].level == 1

        for i in range(0, len(context), 2):
            header, text = context[i], context[i + 1]
            optimized_context.append(
                ContextLevel(
                    level=header.level, header=header.value, content=text.value
                )
            )
        if last_header is not None:
            optimized_context.append(last_header)

        lst = [optimized_context[-1]]
        for i in range(len(optimized_context) - 2, -1, -1):
            if optimized_context[i].level < lst[-1].level:
                lst.append(optimized_context[i])
        return list(reversed(lst))

    def _postprocess_wikipedia(self, table: RemoteExtractedTable):
        for row in table.rows:
            for cell in row.cells:
                for el in cell.travel_elements_post_order():
                    if el.tag == "a":
                        if "href" not in el.attrs:
                            if "selflink" in el.attrs["class"]:
                                el.attrs["href"] = table.page_url
                            else:
                                raise Exception(
                                    f"An anchor in a Wikipedia table does not have href: {el}"
                                )


@M.cache_func()
def requests_get(url):
    return requests.get(url)


def fetch_tables(
    url: str, auto_span: bool = True, auto_pad: bool = True, cache: bool = False
):
    if cache:
        resp = requests_get(url)
    else:
        resp = requests.get(url)

    assert resp.status_code == 200
    html = resp.text
    tables = HTMLTableExtractor().extract(url, html)
    if auto_span:
        new_tables = []
        for tbl in tables:
            try:
                tbl = tbl.span()
                new_tables.append(tbl)
            except OverlapSpanException:
                pass
            except InvalidColumnSpanException:
                pass
        tables = new_tables
    if auto_pad:
        tables = [tbl.pad() for tbl in tables]
    return tables


if __name__ == "__main__":
    # tables = fetch_tables("https://en.wikipedia.org/wiki/President_of_the_National_Council_(Austria)")
    # tables = fetch_tables("https://en.wikipedia.org/wiki/List_of_largest_selling_pharmaceutical_products")
    tables = fetch_tables(
        "https://en.wikipedia.org/wiki/President_of_the_National_Council_(Austria)"
    )
    tables[10].as_relational_linked_table()
    # print(tables[1].as_df())
    # M.serialize_json(tables[6].as_relational_linked_table().to_dict(),
    #                  "/workspace/sm-dev/grams/examples/misc/table_01.json")
